# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.


from pathlib import Path
from typing import Dict, List, Optional

from gems.model.library import Library
from gems.model.parsing import parse_yaml_library
from gems.model.resolve_library import resolve_library
from gems.simulation import TimeBlock, build_problem
from gems.study import DataBase
from gems.study.parsing import parse_cli, parse_yaml_components
from gems.study.resolve_components import (
    System,
    build_data_base,
    build_network,
    consistency_check,
    resolve_system,
)


class AntaresTimeSeriesImportError(Exception):
    pass


def input_libs(yaml_lib_paths: List[Path]) -> dict[str, Library]:
    yaml_libraries = []
    yaml_library_ids = set()

    for path in yaml_lib_paths:
        with path.open("r") as file:
            yaml_lib = parse_yaml_library(file)

            if yaml_lib.id in yaml_library_ids:
                raise ValueError(f"The identifier '{yaml_lib.id}' is defined twice")

            yaml_libraries.append(yaml_lib)
            yaml_library_ids.add(yaml_lib.id)

    return resolve_library(yaml_libraries)


def input_database(study_path: Path, timeseries_path: Optional[Path]) -> DataBase:
    with study_path.open() as comp:
        return build_data_base(parse_yaml_components(comp), timeseries_path)


def input_study(study_path: Path, librairies: dict[str, Library]) -> System:
    with study_path.open() as comp:
        return resolve_system(parse_yaml_components(comp), librairies)


def main_cli() -> None:
    parsed_args = parse_cli()

    lib_dict = input_libs(parsed_args.models_path)
    study = input_study(parsed_args.components_path, lib_dict)

    models = {}
    for lib in lib_dict.values():
        models.update(lib.models)

    consistency_check(study.components, models)

    try:
        database = input_database(
            parsed_args.components_path, parsed_args.timeseries_path
        )

    except UnboundLocalError:
        raise AntaresTimeSeriesImportError(
            f"An error occurred while importing time series."
        )

    network = build_network(study)

    timeblock = TimeBlock(1, list(range(parsed_args.duration)))
    scenario = parsed_args.nb_scenarios

    try:
        problem = build_problem(network, database, timeblock, scenario)

    except IndexError as e:
        raise IndexError(
            f"{e}. Did parameters '--duration' and '--scenario' were correctly set?"
        )

    status = problem.solver.Solve()
    print("status : ", status)

    print("final average cost : ", problem.solver.Objective().Value())


if __name__ == "__main__":
    main_cli()



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict

from gems.expression.expression import (
    AllTimeSumNode,
    ComponentParameterNode,
    ComponentVariableNode,
    PortFieldAggregatorNode,
    PortFieldNode,
    ProblemParameterNode,
    ProblemVariableNode,
    TimeEvalNode,
    TimeShiftNode,
    TimeSumNode,
)

from .expression import (
    ComparisonNode,
    ExpressionNode,
    LiteralNode,
    ParameterNode,
    ScenarioOperatorNode,
    VariableNode,
)
from .indexing import IndexingStructureProvider
from .visitor import ExpressionVisitorOperations, visit


class ValueProvider(ABC):
    """
    Implementations are in charge of mapping parameters and variables to their values.
    Depending on the implementation, evaluation may require a component id or not.
    """

    @abstractmethod
    def get_variable_value(self, name: str) -> float:
        ...

    @abstractmethod
    def get_parameter_value(self, name: str) -> float:
        ...

    @abstractmethod
    def get_component_variable_value(self, component_id: str, name: str) -> float:
        ...

    @abstractmethod
    def get_component_parameter_value(self, component_id: str, name: str) -> float:
        ...


@dataclass(frozen=True)
class EvaluationContext(ValueProvider):
    """
    Simple value provider relying on dictionaries.
    Does not support component variables/parameters.
    """

    variables: Dict[str, float] = field(default_factory=dict)
    parameters: Dict[str, float] = field(default_factory=dict)

    def get_variable_value(self, name: str) -> float:
        return self.variables[name]

    def get_parameter_value(self, name: str) -> float:
        return self.parameters[name]

    def get_component_variable_value(self, component_id: str, name: str) -> float:
        raise NotImplementedError()

    def get_component_parameter_value(self, component_id: str, name: str) -> float:
        raise NotImplementedError()


@dataclass(frozen=True)
class EvaluationVisitor(ExpressionVisitorOperations[float]):
    """
    Evaluates the expression with respect to the provided context
    (variables and parameters values).
    """

    context: ValueProvider

    def literal(self, node: LiteralNode) -> float:
        return node.value

    def comparison(self, node: ComparisonNode) -> float:
        raise ValueError("Cannot evaluate comparison operator.")

    def variable(self, node: VariableNode) -> float:
        return self.context.get_variable_value(node.name)

    def parameter(self, node: ParameterNode) -> float:
        return self.context.get_parameter_value(node.name)

    def comp_parameter(self, node: ComponentParameterNode) -> float:
        return self.context.get_component_parameter_value(node.component_id, node.name)

    def comp_variable(self, node: ComponentVariableNode) -> float:
        return self.context.get_component_variable_value(node.component_id, node.name)

    def pb_parameter(self, node: ProblemParameterNode) -> float:
        raise ValueError("Should not reach here.")

    def pb_variable(self, node: ProblemVariableNode) -> float:
        raise ValueError("Should not reach here.")

    def time_shift(self, node: TimeShiftNode) -> float:
        raise NotImplementedError()

    def time_eval(self, node: TimeEvalNode) -> float:
        raise NotImplementedError()

    def time_sum(self, node: TimeSumNode) -> float:
        raise NotImplementedError()

    def all_time_sum(self, node: AllTimeSumNode) -> float:
        raise NotImplementedError()

    def scenario_operator(self, node: ScenarioOperatorNode) -> float:
        raise NotImplementedError()

    def port_field(self, node: PortFieldNode) -> float:
        raise NotImplementedError()

    def port_field_aggregator(self, node: PortFieldAggregatorNode) -> float:
        raise NotImplementedError()


def evaluate(expression: ExpressionNode, value_provider: ValueProvider) -> float:
    return visit(expression, EvaluationVisitor(value_provider))


class EvaluationError(Exception):
    """Raised when an expression cannot be evaluated due to missing context or math errors."""

    pass



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass, replace
from typing import Any, List

from gems.expression import (
    AdditionNode,
    ComparisonNode,
    DivisionNode,
    ExpressionNode,
    ExpressionVisitor,
    LiteralNode,
    MultiplicationNode,
    NegationNode,
    ParameterNode,
    VariableNode,
)
from gems.expression.expression import (
    AllTimeSumNode,
    BinaryOperatorNode,
    ComponentParameterNode,
    ComponentVariableNode,
    PortFieldAggregatorNode,
    PortFieldNode,
    ProblemParameterNode,
    ProblemVariableNode,
    ScenarioOperatorNode,
    TimeEvalNode,
    TimeShiftNode,
    TimeSumNode,
)
from gems.expression.visitor import visit


@dataclass(frozen=True)
class PortField:
    name: str


@dataclass
class PortType:
    """
    Defines a port type.

    A port is an external interface of a model, where other
    ports can be connected.
    Only compatible ports may be connected together (?)
    """

    id: str
    fields: List[PortField]  # TODO: should we rename with "pin" ?


@dataclass(frozen=True)
class PortFieldId:
    port_name: str
    field_name: str

    def replicate(self, /, **changes: Any) -> "PortFieldId":
        return replace(self, **changes)


@dataclass(frozen=True)
class PortFieldDefinition:
    """
    Defines the values of one port field
    """

    port_field: PortFieldId
    definition: ExpressionNode

    def __post_init__(self) -> None:
        _validate_port_field_expression(self)

    def replicate(self, /, **changes: Any) -> "PortFieldDefinition":
        return replace(self, **changes)


def port_field_def(
    port_name: str, field_name: str, definition: ExpressionNode
) -> PortFieldDefinition:
    return PortFieldDefinition(PortFieldId(port_name, field_name), definition)


class _PortFieldExpressionChecker(ExpressionVisitor[None]):
    """
    Visits the whole expression to check there is no:
    comparison, other port field, component-associated parametrs or variables...
    """

    def literal(self, node: LiteralNode) -> None:
        pass

    def negation(self, node: NegationNode) -> None:
        visit(node.operand, self)

    def _visit_binary_op(self, node: BinaryOperatorNode) -> None:
        visit(node.left, self)
        visit(node.right, self)

    def addition(self, node: AdditionNode) -> None:
        for n in node.operands:
            visit(n, self)

    def multiplication(self, node: MultiplicationNode) -> None:
        self._visit_binary_op(node)

    def division(self, node: DivisionNode) -> None:
        self._visit_binary_op(node)

    def comparison(self, node: ComparisonNode) -> None:
        raise ValueError("Port definition cannot contain a comparison operator.")

    def variable(self, node: VariableNode) -> None:
        pass

    def parameter(self, node: ParameterNode) -> None:
        pass

    def comp_parameter(self, node: ComponentParameterNode) -> None:
        raise ValueError(
            "Port definition must not contain a parameter associated to a component."
        )

    def comp_variable(self, node: ComponentVariableNode) -> None:
        raise ValueError(
            "Port definition must not contain a variable associated to a component."
        )

    def pb_parameter(self, node: ProblemParameterNode) -> None:
        raise ValueError(
            "Port definition must not contain a parameter associated to a component."
        )

    def pb_variable(self, node: ProblemVariableNode) -> None:
        raise ValueError(
            "Port definition must not contain a variable associated to a component."
        )

    def time_shift(self, node: TimeShiftNode) -> None:
        visit(node.operand, self)

    def time_eval(self, node: TimeEvalNode) -> None:
        visit(node.operand, self)

    def time_sum(self, node: TimeSumNode) -> None:
        visit(node.operand, self)

    def all_time_sum(self, node: AllTimeSumNode) -> None:
        visit(node.operand, self)

    def scenario_operator(self, node: ScenarioOperatorNode) -> None:
        visit(node.operand, self)

    def port_field(self, node: PortFieldNode) -> None:
        raise ValueError("Port definition cannot reference another port field.")

    def port_field_aggregator(self, node: PortFieldAggregatorNode) -> None:
        raise ValueError("Port definition cannot contain port field aggregation.")


def _validate_port_field_expression(definition: PortFieldDefinition) -> None:
    visit(definition.definition, _PortFieldExpressionChecker())



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from .data import (
    ConstantData,
    DataBase,
    ScenarioIndex,
    ScenarioSeriesData,
    Scenarization,
    TimeIndex,
    TimeScenarioIndex,
    TimeScenarioSeriesData,
    TimeSeriesData,
)
from .network import (
    Component,
    Network,
    Node,
    PortRef,
    PortsConnection,
    create_component,
)



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
import typing
from dataclasses import dataclass
from typing import List, Optional

from pydantic import Field, ValidationError
from yaml import safe_load

from gems.utils import ModifiedBaseModel


def parse_yaml_library(input: typing.TextIO) -> "InputLibrary":
    tree = safe_load(input)
    try:
        return InputLibrary.model_validate(tree["library"])
    except ValidationError as e:
        raise ValueError(f"An error occurred during parsing: {e}")


class InputParameter(ModifiedBaseModel):
    id: str
    time_dependent: bool = False
    scenario_dependent: bool = False


class InputVariable(ModifiedBaseModel):
    id: str
    time_dependent: bool = True
    scenario_dependent: bool = True
    lower_bound: Optional[str] = None
    upper_bound: Optional[str] = None
    variable_type: str = "continuous"

    class Config(ModifiedBaseModel.Config):
        coerce_numbers_to_str = True


class InputConstraint(ModifiedBaseModel):
    id: str
    expression: str
    lower_bound: Optional[str] = None
    upper_bound: Optional[str] = None


class InputField(ModifiedBaseModel):
    id: str


class InputPortType(ModifiedBaseModel):
    id: str
    fields: List[InputField] = Field(default_factory=list)
    description: Optional[str] = None


class InputModelPort(ModifiedBaseModel):
    id: str
    type: str


class InputPortFieldDefinition(ModifiedBaseModel):
    port: str
    field: str
    definition: str


class InputObjectiveContribution(ModifiedBaseModel):
    id: str
    expression: str


@dataclass
class InputExtraOutput(ModifiedBaseModel):
    id: str
    expression: str


class InputModel(ModifiedBaseModel):
    id: str
    parameters: List[InputParameter] = Field(default_factory=list)
    variables: List[InputVariable] = Field(default_factory=list)
    ports: List[InputModelPort] = Field(default_factory=list)
    port_field_definitions: List[InputPortFieldDefinition] = Field(default_factory=list)
    binding_constraints: List[InputConstraint] = Field(default_factory=list)
    constraints: List[InputConstraint] = Field(default_factory=list)
    objective_contributions: List[InputObjectiveContribution] = Field(
        default_factory=list, alias="objective-contributions"
    )
    description: Optional[str] = None
    extra_outputs: Optional[List[InputExtraOutput]] = None


class InputLibrary(ModifiedBaseModel):
    id: str
    dependencies: List[str] = Field(default_factory=list)
    port_types: List[InputPortType] = Field(default_factory=list)
    models: List[InputModel] = Field(default_factory=list)
    description: Optional[str] = None



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
Defines the model for generic expressions.
"""
import enum
import inspect
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Sequence, Union

import gems.expression.port_operator
import gems.expression.scenario_operator

AnyExpression = Union[int, float, "ExpressionNode"]


@dataclass(frozen=True)
class ExpressionNode:
    """
    Base class for all nodes of the expression AST.

    Operators overloading is provided to help create expressions
    programmatically.

    Examples
        >>> expr = -var('x') + 5 / param('p')
    """

    def __neg__(self) -> "ExpressionNode":
        return NegationNode(self)

    def __add__(self, rhs: Any) -> "ExpressionNode":
        lhs = self
        operands = []
        rhs = _wrap_in_node(rhs)
        operands.extend(lhs.operands if isinstance(lhs, AdditionNode) else [lhs])
        operands.extend(rhs.operands if isinstance(rhs, AdditionNode) else [rhs])
        return AdditionNode(operands)

    def __radd__(self, lhs: Any) -> "ExpressionNode":
        lhs = _wrap_in_node(lhs)
        return lhs + self

    def __sub__(self, rhs: Any) -> "ExpressionNode":
        lhs = self
        operands = []
        rhs = _wrap_in_node(rhs)
        operands.extend(lhs.operands if isinstance(lhs, AdditionNode) else [lhs])
        right_operands = rhs.operands if isinstance(rhs, AdditionNode) else [rhs]
        operands.extend([-o for o in right_operands])
        return AdditionNode(operands)

    def __rsub__(self, lhs: Any) -> "ExpressionNode":
        lhs = _wrap_in_node(lhs)
        return lhs + self

    def __mul__(self, rhs: Any) -> "ExpressionNode":
        return _apply_if_node(rhs, lambda x: MultiplicationNode(self, x))

    def __rmul__(self, lhs: Any) -> "ExpressionNode":
        return _apply_if_node(lhs, lambda x: MultiplicationNode(x, self))

    def __truediv__(self, rhs: Any) -> "ExpressionNode":
        return _apply_if_node(rhs, lambda x: DivisionNode(self, x))

    def __rtruediv__(self, lhs: Any) -> "ExpressionNode":
        return _apply_if_node(lhs, lambda x: DivisionNode(x, self))

    def __le__(self, rhs: Any) -> "ExpressionNode":
        return _apply_if_node(
            rhs, lambda x: ComparisonNode(self, x, Comparator.LESS_THAN)
        )

    def __ge__(self, rhs: Any) -> "ExpressionNode":
        return _apply_if_node(
            rhs, lambda x: ComparisonNode(self, x, Comparator.GREATER_THAN)
        )

    def __eq__(self, rhs: Any) -> "ExpressionNode":  # type: ignore
        return _apply_if_node(rhs, lambda x: ComparisonNode(self, x, Comparator.EQUAL))

    def time_sum(
        self,
        from_shift: Optional[AnyExpression] = None,
        to_shift: Optional[AnyExpression] = None,
    ) -> "ExpressionNode":
        if from_shift is None and to_shift is None:
            return AllTimeSumNode(self)
        if from_shift is None or to_shift is None:
            raise ValueError("Both time bounds of a time sum must be defined.")
        return TimeSumNode(
            operand=self,
            from_time=_wrap_in_node(from_shift),
            to_time=_wrap_in_node(to_shift),
        )

    def sum_connections(self) -> "ExpressionNode":
        if isinstance(self, PortFieldNode):
            return PortFieldAggregatorNode(self, aggregator="PortSum")
        raise ValueError(
            f"sum_connections() applies only for PortFieldNode, whereas the current node is of type {type(self)}."
        )

    def shift(self, shift: AnyExpression) -> "ExpressionNode":
        return TimeShiftNode(self, _wrap_in_node(shift))

    def eval(self, time: AnyExpression) -> "ExpressionNode":
        return TimeEvalNode(self, _wrap_in_node(time))

    def expec(self) -> "ExpressionNode":
        return _apply_if_node(self, lambda x: ScenarioOperatorNode(x, "Expectation"))

    def variance(self) -> "ExpressionNode":
        return _apply_if_node(self, lambda x: ScenarioOperatorNode(x, "Variance"))


def _wrap_in_node(obj: Any) -> ExpressionNode:
    if isinstance(obj, ExpressionNode):
        return obj
    elif isinstance(obj, float) or isinstance(obj, int):
        return LiteralNode(float(obj))
    raise TypeError(f"Unable to wrap {obj} into an expression node")


def _apply_if_node(
    obj: Any, func: Callable[["ExpressionNode"], "ExpressionNode"]
) -> "ExpressionNode":
    if as_node := _wrap_in_node(obj):
        return func(as_node)
    else:
        return NotImplemented


@dataclass(frozen=True, eq=False)
class VariableNode(ExpressionNode):
    name: str


def var(name: str) -> VariableNode:
    return VariableNode(name)


@dataclass(frozen=True, eq=False)
class PortFieldNode(ExpressionNode):
    """
    References a port field.
    """

    port_name: str
    field_name: str


def port_field(port_name: str, field_name: str) -> PortFieldNode:
    return PortFieldNode(port_name, field_name)


@dataclass(frozen=True, eq=False)
class ParameterNode(ExpressionNode):
    name: str


def param(name: str) -> ParameterNode:
    return ParameterNode(name)


@dataclass(frozen=True, eq=False)
class ComponentParameterNode(ExpressionNode):
    """
    Represents one parameter of one component.

    When building actual equations for a system,
    we need to associated each parameter to its
    actual component, at some point.
    """

    component_id: str
    name: str


def comp_param(component_id: str, name: str) -> ComponentParameterNode:
    return ComponentParameterNode(component_id, name)


@dataclass(frozen=True)
class TimeIndex:
    pass


@dataclass(frozen=True)
class NoTimeIndex(TimeIndex):
    """
    Some values do not depend on the timestep, this index should be used for them
    (think one variable for all timesteps).
    """

    pass


@dataclass(frozen=True)
class TimeShift(TimeIndex):
    """
    Represents the current timestep + a shift.

    This should only be used for nodes that actually depend on the timestep,
    never for time-independent nodes (constant parameters ...).
    """

    timeshift: int


@dataclass(frozen=True)
class TimeStep(TimeIndex):
    """
    Represents a given timestep, independently of the current timestep.

    This should only be used for nodes that actually depend on the timestep,
    never for time-independent nodes (constant parameters ...).
    """

    timestep: int


@dataclass(frozen=True)
class ScenarioIndex:
    pass


@dataclass(frozen=True)
class NoScenarioIndex(ScenarioIndex):
    """
    Some values do not depend on the scenario, this index should be used for them
    (think one variable for all timesteps).
    """

    pass


@dataclass(frozen=True)
class CurrentScenarioIndex(ScenarioIndex):
    """
    Represents the current scenario.

    This should only be used for nodes that actually depend on the scenario,
    never for scenario-independent nodes (constant parameters ...).
    """

    pass


@dataclass(frozen=True)
class OneScenarioIndex(ScenarioIndex):
    """
    Represents a given scenario out of all scenarios.

    This should only be used for nodes that actually depend on the scenario,
    never for scenario-independent nodes (constant parameters ...).
    """

    scenario: int


@dataclass(frozen=True, eq=False)
class ProblemParameterNode(ExpressionNode):
    """
    Represents one parameter of the optimization problem
    """

    component_id: str
    name: str
    time_index: TimeIndex
    scenario_index: ScenarioIndex


def problem_param(
    component_id: str, name: str, time_index: TimeIndex, scenario_index: ScenarioIndex
) -> ProblemParameterNode:
    return ProblemParameterNode(component_id, name, time_index, scenario_index)


@dataclass(frozen=True, eq=False)
class ComponentVariableNode(ExpressionNode):
    """
    Represents one variable of one component.

    When building actual equations for a system,
    we need to associated each variable to its
    actual component, at some point.
    """

    component_id: str
    name: str


def comp_var(component_id: str, name: str) -> ComponentVariableNode:
    return ComponentVariableNode(component_id, name)


@dataclass(frozen=True, eq=False)
class ProblemVariableNode(ExpressionNode):
    """
    Represents one variable of the optimization problem
    """

    component_id: str
    name: str
    time_index: TimeIndex
    scenario_index: ScenarioIndex


def problem_var(
    component_id: str, name: str, time_index: TimeIndex, scenario_index: ScenarioIndex
) -> ProblemVariableNode:
    return ProblemVariableNode(component_id, name, time_index, scenario_index)


@dataclass(frozen=True, eq=False)
class LiteralNode(ExpressionNode):
    value: float


def literal(value: float) -> LiteralNode:
    return LiteralNode(value)


def is_unbounded(expr: ExpressionNode) -> bool:
    return isinstance(expr, LiteralNode) and (abs(expr.value) == float("inf"))


def is_non_negative(expr: ExpressionNode) -> bool:
    return isinstance(expr, LiteralNode) and (expr.value >= 0)


@dataclass(frozen=True, eq=False)
class UnaryOperatorNode(ExpressionNode):
    operand: ExpressionNode


@dataclass(frozen=True, eq=False)
class PortFieldAggregatorNode(UnaryOperatorNode):
    aggregator: str

    def __post_init__(self) -> None:
        valid_names = [
            cls.__name__
            for _, cls in inspect.getmembers(
                gems.expression.port_operator, inspect.isclass
            )
            if issubclass(cls, gems.expression.port_operator.PortAggregator)
        ]
        if self.aggregator not in valid_names:
            raise NotImplementedError(
                f"{self.aggregator} is not a valid port aggregator, valid port aggregators are {valid_names}"
            )


@dataclass(frozen=True, eq=False)
class NegationNode(UnaryOperatorNode):
    pass


@dataclass(frozen=True, eq=False)
class BinaryOperatorNode(ExpressionNode):
    left: ExpressionNode
    right: ExpressionNode


class Comparator(enum.Enum):
    LESS_THAN = "LESS_THAN"
    EQUAL = "EQUAL"
    GREATER_THAN = "GREATER_THAN"


@dataclass(frozen=True, eq=False)
class ComparisonNode(BinaryOperatorNode):
    comparator: Comparator


@dataclass(frozen=True, eq=False)
class AdditionNode(ExpressionNode):
    operands: List[ExpressionNode]


@dataclass(frozen=True, eq=False)
class MultiplicationNode(BinaryOperatorNode):
    pass


@dataclass(frozen=True, eq=False)
class DivisionNode(BinaryOperatorNode):
    pass


@dataclass(frozen=True, eq=False)
class TimeShiftNode(UnaryOperatorNode):
    time_shift: ExpressionNode


@dataclass(frozen=True, eq=False)
class TimeEvalNode(UnaryOperatorNode):
    eval_time: ExpressionNode


@dataclass(frozen=True, eq=False)
class TimeSumNode(UnaryOperatorNode):
    from_time: ExpressionNode
    to_time: ExpressionNode


@dataclass(frozen=True, eq=False)
class AllTimeSumNode(UnaryOperatorNode):
    """
    Separate from time sum node because it's actually a quite different operation:
    In particular, this changes the time indexing.
    """

    pass


@dataclass(frozen=True, eq=False)
class ScenarioOperatorNode(UnaryOperatorNode):
    name: str

    def __post_init__(self) -> None:
        valid_names = [
            cls.__name__
            for _, cls in inspect.getmembers(
                gems.expression.scenario_operator, inspect.isclass
            )
            if issubclass(cls, gems.expression.scenario_operator.ScenarioOperator)
        ]
        if self.name not in valid_names:
            raise ValueError(
                f"{self.name} is not a valid scenario operator, valid scenario operators are {valid_names}"
            )


def sum_expressions(expressions: Sequence[ExpressionNode]) -> ExpressionNode:
    if len(expressions) == 0:
        return LiteralNode(0)
    if len(expressions) == 1:
        return expressions[0]
    return AdditionNode([e for e in expressions])



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
Defines abstract base class for visitors of expressions.
"""
import typing
from abc import ABC, abstractmethod
from typing import Generic, Protocol, TypeVar

from gems.expression.expression import (
    AdditionNode,
    AllTimeSumNode,
    ComparisonNode,
    ComponentParameterNode,
    ComponentVariableNode,
    DivisionNode,
    ExpressionNode,
    LiteralNode,
    MultiplicationNode,
    NegationNode,
    ParameterNode,
    PortFieldAggregatorNode,
    PortFieldNode,
    ProblemParameterNode,
    ProblemVariableNode,
    ScenarioOperatorNode,
    TimeEvalNode,
    TimeShiftNode,
    TimeSumNode,
    VariableNode,
)

T = TypeVar("T")


class ExpressionVisitor(ABC, Generic[T]):
    """
    Abstract base class for visitors of expressions.

    Visitor can be implemented to carry out arbitrary kind
    of analysis over an expression, such as evaluation,
    manipulation ...
    """

    @abstractmethod
    def literal(self, node: LiteralNode) -> T:
        ...

    @abstractmethod
    def negation(self, node: NegationNode) -> T:
        ...

    @abstractmethod
    def addition(self, node: AdditionNode) -> T:
        ...

    @abstractmethod
    def multiplication(self, node: MultiplicationNode) -> T:
        ...

    @abstractmethod
    def division(self, node: DivisionNode) -> T:
        ...

    @abstractmethod
    def comparison(self, node: ComparisonNode) -> T:
        ...

    @abstractmethod
    def variable(self, node: VariableNode) -> T:
        ...

    @abstractmethod
    def parameter(self, node: ParameterNode) -> T:
        ...

    @abstractmethod
    def comp_parameter(self, node: ComponentParameterNode) -> T:
        ...

    @abstractmethod
    def comp_variable(self, node: ComponentVariableNode) -> T:
        ...

    @abstractmethod
    def pb_parameter(self, node: ProblemParameterNode) -> T:
        ...

    @abstractmethod
    def pb_variable(self, node: ProblemVariableNode) -> T:
        ...

    @abstractmethod
    def time_shift(self, node: TimeShiftNode) -> T:
        ...

    @abstractmethod
    def time_eval(self, node: TimeEvalNode) -> T:
        ...

    @abstractmethod
    def time_sum(self, node: TimeSumNode) -> T:
        ...

    @abstractmethod
    def all_time_sum(self, node: AllTimeSumNode) -> T:
        ...

    @abstractmethod
    def scenario_operator(self, node: ScenarioOperatorNode) -> T:
        ...

    @abstractmethod
    def port_field(self, node: PortFieldNode) -> T:
        ...

    @abstractmethod
    def port_field_aggregator(self, node: PortFieldAggregatorNode) -> T:
        ...


def visit(root: ExpressionNode, visitor: ExpressionVisitor[T]) -> T:
    """
    Utility method to dispatch calls to the right method of a visitor.
    """
    if isinstance(root, LiteralNode):
        return visitor.literal(root)
    elif isinstance(root, NegationNode):
        return visitor.negation(root)
    elif isinstance(root, VariableNode):
        return visitor.variable(root)
    elif isinstance(root, ParameterNode):
        return visitor.parameter(root)
    elif isinstance(root, ComponentParameterNode):
        return visitor.comp_parameter(root)
    elif isinstance(root, ComponentVariableNode):
        return visitor.comp_variable(root)
    elif isinstance(root, ProblemParameterNode):
        return visitor.pb_parameter(root)
    elif isinstance(root, ProblemVariableNode):
        return visitor.pb_variable(root)
    elif isinstance(root, AdditionNode):
        return visitor.addition(root)
    elif isinstance(root, MultiplicationNode):
        return visitor.multiplication(root)
    elif isinstance(root, DivisionNode):
        return visitor.division(root)
    elif isinstance(root, ComparisonNode):
        return visitor.comparison(root)
    elif isinstance(root, TimeShiftNode):
        return visitor.time_shift(root)
    elif isinstance(root, TimeEvalNode):
        return visitor.time_eval(root)
    elif isinstance(root, TimeSumNode):
        return visitor.time_sum(root)
    elif isinstance(root, AllTimeSumNode):
        return visitor.all_time_sum(root)
    elif isinstance(root, ScenarioOperatorNode):
        return visitor.scenario_operator(root)
    elif isinstance(root, PortFieldNode):
        return visitor.port_field(root)
    elif isinstance(root, PortFieldAggregatorNode):
        return visitor.port_field_aggregator(root)
    raise ValueError(f"Unknown expression node type {root.__class__}")


class SupportsOperations(Protocol[T]):
    """
    Defines a type which implements math operations +, -, *, /
    """

    @abstractmethod
    def __add__(self, other: T) -> T:
        pass

    @abstractmethod
    def __neg__(self) -> T:
        pass

    @abstractmethod
    def __sub__(self, other: T) -> T:
        pass

    @abstractmethod
    def __mul__(self, other: T) -> T:
        pass

    @abstractmethod
    def __truediv__(self, other: T) -> T:
        pass


T_op = TypeVar("T_op", bound=SupportsOperations)


class ExpressionVisitorOperations(ExpressionVisitor[T_op], ABC):
    """
    Provides default implementations of math operations
    based on (+, -, /, *) operations of type T.
    """

    def negation(self, node: NegationNode) -> T_op:
        return -visit(node.operand, self)

    def addition(self, node: AdditionNode) -> T_op:
        operands = [visit(o, self) for o in node.operands]
        res = operands[0]
        for o in operands[1:]:
            res = res + o
        return res

    def multiplication(self, node: MultiplicationNode) -> T_op:
        left_value = visit(node.left, self)
        right_value = visit(node.right, self)
        return left_value * right_value

    def division(self, node: DivisionNode) -> T_op:
        left_value = visit(node.left, self)
        right_value = visit(node.right, self)
        return left_value / right_value



# Generated from Expr.g4 by ANTLR 4.13.2
# encoding: utf-8
import sys
from io import StringIO

from antlr4 import *

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    return [
        4,
        1,
        18,
        140,
        2,
        0,
        7,
        0,
        2,
        1,
        7,
        1,
        2,
        2,
        7,
        2,
        2,
        3,
        7,
        3,
        2,
        4,
        7,
        4,
        2,
        5,
        7,
        5,
        2,
        6,
        7,
        6,
        1,
        0,
        1,
        0,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        3,
        2,
        79,
        8,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        5,
        2,
        90,
        8,
        2,
        10,
        2,
        12,
        2,
        93,
        9,
        2,
        1,
        3,
        1,
        3,
        3,
        3,
        97,
        8,
        3,
        1,
        4,
        1,
        4,
        3,
        4,
        101,
        8,
        4,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        3,
        5,
        111,
        8,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        5,
        5,
        119,
        8,
        5,
        10,
        5,
        12,
        5,
        122,
        9,
        5,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        3,
        6,
        130,
        8,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        5,
        6,
        135,
        8,
        6,
        10,
        6,
        12,
        6,
        138,
        9,
        6,
        1,
        6,
        0,
        3,
        4,
        10,
        12,
        7,
        0,
        2,
        4,
        6,
        8,
        10,
        12,
        0,
        2,
        1,
        0,
        5,
        6,
        2,
        0,
        2,
        2,
        7,
        7,
        153,
        0,
        14,
        1,
        0,
        0,
        0,
        2,
        18,
        1,
        0,
        0,
        0,
        4,
        78,
        1,
        0,
        0,
        0,
        6,
        96,
        1,
        0,
        0,
        0,
        8,
        98,
        1,
        0,
        0,
        0,
        10,
        110,
        1,
        0,
        0,
        0,
        12,
        129,
        1,
        0,
        0,
        0,
        14,
        15,
        5,
        16,
        0,
        0,
        15,
        16,
        5,
        1,
        0,
        0,
        16,
        17,
        5,
        16,
        0,
        0,
        17,
        1,
        1,
        0,
        0,
        0,
        18,
        19,
        3,
        4,
        2,
        0,
        19,
        20,
        5,
        0,
        0,
        1,
        20,
        3,
        1,
        0,
        0,
        0,
        21,
        22,
        6,
        2,
        -1,
        0,
        22,
        79,
        3,
        6,
        3,
        0,
        23,
        79,
        3,
        0,
        0,
        0,
        24,
        25,
        5,
        2,
        0,
        0,
        25,
        79,
        3,
        4,
        2,
        13,
        26,
        27,
        5,
        3,
        0,
        0,
        27,
        28,
        3,
        4,
        2,
        0,
        28,
        29,
        5,
        4,
        0,
        0,
        29,
        79,
        1,
        0,
        0,
        0,
        30,
        31,
        5,
        8,
        0,
        0,
        31,
        32,
        5,
        3,
        0,
        0,
        32,
        33,
        3,
        4,
        2,
        0,
        33,
        34,
        5,
        4,
        0,
        0,
        34,
        79,
        1,
        0,
        0,
        0,
        35,
        36,
        5,
        9,
        0,
        0,
        36,
        37,
        5,
        3,
        0,
        0,
        37,
        38,
        3,
        0,
        0,
        0,
        38,
        39,
        5,
        4,
        0,
        0,
        39,
        79,
        1,
        0,
        0,
        0,
        40,
        41,
        5,
        8,
        0,
        0,
        41,
        42,
        5,
        3,
        0,
        0,
        42,
        43,
        3,
        8,
        4,
        0,
        43,
        44,
        5,
        10,
        0,
        0,
        44,
        45,
        3,
        8,
        4,
        0,
        45,
        46,
        5,
        11,
        0,
        0,
        46,
        47,
        3,
        4,
        2,
        0,
        47,
        48,
        5,
        4,
        0,
        0,
        48,
        79,
        1,
        0,
        0,
        0,
        49,
        50,
        5,
        16,
        0,
        0,
        50,
        51,
        5,
        3,
        0,
        0,
        51,
        52,
        3,
        4,
        2,
        0,
        52,
        53,
        5,
        4,
        0,
        0,
        53,
        79,
        1,
        0,
        0,
        0,
        54,
        55,
        5,
        16,
        0,
        0,
        55,
        56,
        5,
        12,
        0,
        0,
        56,
        57,
        3,
        8,
        4,
        0,
        57,
        58,
        5,
        13,
        0,
        0,
        58,
        79,
        1,
        0,
        0,
        0,
        59,
        60,
        5,
        16,
        0,
        0,
        60,
        61,
        5,
        12,
        0,
        0,
        61,
        62,
        3,
        4,
        2,
        0,
        62,
        63,
        5,
        13,
        0,
        0,
        63,
        79,
        1,
        0,
        0,
        0,
        64,
        65,
        5,
        3,
        0,
        0,
        65,
        66,
        3,
        4,
        2,
        0,
        66,
        67,
        5,
        4,
        0,
        0,
        67,
        68,
        5,
        12,
        0,
        0,
        68,
        69,
        3,
        8,
        4,
        0,
        69,
        70,
        5,
        13,
        0,
        0,
        70,
        79,
        1,
        0,
        0,
        0,
        71,
        72,
        5,
        3,
        0,
        0,
        72,
        73,
        3,
        4,
        2,
        0,
        73,
        74,
        5,
        4,
        0,
        0,
        74,
        75,
        5,
        12,
        0,
        0,
        75,
        76,
        3,
        4,
        2,
        0,
        76,
        77,
        5,
        13,
        0,
        0,
        77,
        79,
        1,
        0,
        0,
        0,
        78,
        21,
        1,
        0,
        0,
        0,
        78,
        23,
        1,
        0,
        0,
        0,
        78,
        24,
        1,
        0,
        0,
        0,
        78,
        26,
        1,
        0,
        0,
        0,
        78,
        30,
        1,
        0,
        0,
        0,
        78,
        35,
        1,
        0,
        0,
        0,
        78,
        40,
        1,
        0,
        0,
        0,
        78,
        49,
        1,
        0,
        0,
        0,
        78,
        54,
        1,
        0,
        0,
        0,
        78,
        59,
        1,
        0,
        0,
        0,
        78,
        64,
        1,
        0,
        0,
        0,
        78,
        71,
        1,
        0,
        0,
        0,
        79,
        91,
        1,
        0,
        0,
        0,
        80,
        81,
        10,
        11,
        0,
        0,
        81,
        82,
        7,
        0,
        0,
        0,
        82,
        90,
        3,
        4,
        2,
        12,
        83,
        84,
        10,
        10,
        0,
        0,
        84,
        85,
        7,
        1,
        0,
        0,
        85,
        90,
        3,
        4,
        2,
        11,
        86,
        87,
        10,
        9,
        0,
        0,
        87,
        88,
        5,
        17,
        0,
        0,
        88,
        90,
        3,
        4,
        2,
        10,
        89,
        80,
        1,
        0,
        0,
        0,
        89,
        83,
        1,
        0,
        0,
        0,
        89,
        86,
        1,
        0,
        0,
        0,
        90,
        93,
        1,
        0,
        0,
        0,
        91,
        89,
        1,
        0,
        0,
        0,
        91,
        92,
        1,
        0,
        0,
        0,
        92,
        5,
        1,
        0,
        0,
        0,
        93,
        91,
        1,
        0,
        0,
        0,
        94,
        97,
        5,
        14,
        0,
        0,
        95,
        97,
        5,
        16,
        0,
        0,
        96,
        94,
        1,
        0,
        0,
        0,
        96,
        95,
        1,
        0,
        0,
        0,
        97,
        7,
        1,
        0,
        0,
        0,
        98,
        100,
        5,
        15,
        0,
        0,
        99,
        101,
        3,
        10,
        5,
        0,
        100,
        99,
        1,
        0,
        0,
        0,
        100,
        101,
        1,
        0,
        0,
        0,
        101,
        9,
        1,
        0,
        0,
        0,
        102,
        103,
        6,
        5,
        -1,
        0,
        103,
        104,
        7,
        1,
        0,
        0,
        104,
        111,
        3,
        6,
        3,
        0,
        105,
        106,
        7,
        1,
        0,
        0,
        106,
        107,
        5,
        3,
        0,
        0,
        107,
        108,
        3,
        4,
        2,
        0,
        108,
        109,
        5,
        4,
        0,
        0,
        109,
        111,
        1,
        0,
        0,
        0,
        110,
        102,
        1,
        0,
        0,
        0,
        110,
        105,
        1,
        0,
        0,
        0,
        111,
        120,
        1,
        0,
        0,
        0,
        112,
        113,
        10,
        4,
        0,
        0,
        113,
        114,
        7,
        0,
        0,
        0,
        114,
        119,
        3,
        12,
        6,
        0,
        115,
        116,
        10,
        3,
        0,
        0,
        116,
        117,
        7,
        1,
        0,
        0,
        117,
        119,
        3,
        12,
        6,
        0,
        118,
        112,
        1,
        0,
        0,
        0,
        118,
        115,
        1,
        0,
        0,
        0,
        119,
        122,
        1,
        0,
        0,
        0,
        120,
        118,
        1,
        0,
        0,
        0,
        120,
        121,
        1,
        0,
        0,
        0,
        121,
        11,
        1,
        0,
        0,
        0,
        122,
        120,
        1,
        0,
        0,
        0,
        123,
        124,
        6,
        6,
        -1,
        0,
        124,
        125,
        5,
        3,
        0,
        0,
        125,
        126,
        3,
        4,
        2,
        0,
        126,
        127,
        5,
        4,
        0,
        0,
        127,
        130,
        1,
        0,
        0,
        0,
        128,
        130,
        3,
        6,
        3,
        0,
        129,
        123,
        1,
        0,
        0,
        0,
        129,
        128,
        1,
        0,
        0,
        0,
        130,
        136,
        1,
        0,
        0,
        0,
        131,
        132,
        10,
        3,
        0,
        0,
        132,
        133,
        7,
        0,
        0,
        0,
        133,
        135,
        3,
        12,
        6,
        4,
        134,
        131,
        1,
        0,
        0,
        0,
        135,
        138,
        1,
        0,
        0,
        0,
        136,
        134,
        1,
        0,
        0,
        0,
        136,
        137,
        1,
        0,
        0,
        0,
        137,
        13,
        1,
        0,
        0,
        0,
        138,
        136,
        1,
        0,
        0,
        0,
        10,
        78,
        89,
        91,
        96,
        100,
        110,
        118,
        120,
        129,
        136,
    ]


class ExprParser(Parser):
    grammarFileName = "Expr.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "'.'",
        "'-'",
        "'('",
        "')'",
        "'/'",
        "'*'",
        "'+'",
        "'sum'",
        "'sum_connections'",
        "'..'",
        "','",
        "'['",
        "']'",
        "<INVALID>",
        "'t'",
    ]

    symbolicNames = [
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "NUMBER",
        "TIME",
        "IDENTIFIER",
        "COMPARISON",
        "WS",
    ]

    RULE_portFieldExpr = 0
    RULE_fullexpr = 1
    RULE_expr = 2
    RULE_atom = 3
    RULE_shift = 4
    RULE_shift_expr = 5
    RULE_right_expr = 6

    ruleNames = [
        "portFieldExpr",
        "fullexpr",
        "expr",
        "atom",
        "shift",
        "shift_expr",
        "right_expr",
    ]

    EOF = Token.EOF
    T__0 = 1
    T__1 = 2
    T__2 = 3
    T__3 = 4
    T__4 = 5
    T__5 = 6
    T__6 = 7
    T__7 = 8
    T__8 = 9
    T__9 = 10
    T__10 = 11
    T__11 = 12
    T__12 = 13
    NUMBER = 14
    TIME = 15
    IDENTIFIER = 16
    COMPARISON = 17
    WS = 18

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class PortFieldExprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self, i: int = None):
            if i is None:
                return self.getTokens(ExprParser.IDENTIFIER)
            else:
                return self.getToken(ExprParser.IDENTIFIER, i)

        def getRuleIndex(self):
            return ExprParser.RULE_portFieldExpr

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPortFieldExpr"):
                return visitor.visitPortFieldExpr(self)
            else:
                return visitor.visitChildren(self)

    def portFieldExpr(self):
        localctx = ExprParser.PortFieldExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_portFieldExpr)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 14
            self.match(ExprParser.IDENTIFIER)
            self.state = 15
            self.match(ExprParser.T__0)
            self.state = 16
            self.match(ExprParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FullexprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def EOF(self):
            return self.getToken(ExprParser.EOF, 0)

        def getRuleIndex(self):
            return ExprParser.RULE_fullexpr

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFullexpr"):
                return visitor.visitFullexpr(self)
            else:
                return visitor.visitChildren(self)

    def fullexpr(self):
        localctx = ExprParser.FullexprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_fullexpr)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 18
            self.expr(0)
            self.state = 19
            self.match(ExprParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ExprParser.RULE_expr

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class PortFieldSumContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def portFieldExpr(self):
            return self.getTypedRuleContext(ExprParser.PortFieldExprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPortFieldSum"):
                return visitor.visitPortFieldSum(self)
            else:
                return visitor.visitChildren(self)

    class NegationContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitNegation"):
                return visitor.visitNegation(self)
            else:
                return visitor.visitChildren(self)

    class UnsignedAtomContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def atom(self):
            return self.getTypedRuleContext(ExprParser.AtomContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitUnsignedAtom"):
                return visitor.visitUnsignedAtom(self)
            else:
                return visitor.visitChildren(self)

    class ExpressionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitExpression"):
                return visitor.visitExpression(self)
            else:
                return visitor.visitChildren(self)

    class ComparisonContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ExprParser.ExprContext)
            else:
                return self.getTypedRuleContext(ExprParser.ExprContext, i)

        def COMPARISON(self):
            return self.getToken(ExprParser.COMPARISON, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitComparison"):
                return visitor.visitComparison(self)
            else:
                return visitor.visitChildren(self)

    class AllTimeSumContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAllTimeSum"):
                return visitor.visitAllTimeSum(self)
            else:
                return visitor.visitChildren(self)

    class TimeIndexExprContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ExprParser.ExprContext)
            else:
                return self.getTypedRuleContext(ExprParser.ExprContext, i)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTimeIndexExpr"):
                return visitor.visitTimeIndexExpr(self)
            else:
                return visitor.visitChildren(self)

    class AddsubContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.op = None  # Token
            self.copyFrom(ctx)

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ExprParser.ExprContext)
            else:
                return self.getTypedRuleContext(ExprParser.ExprContext, i)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAddsub"):
                return visitor.visitAddsub(self)
            else:
                return visitor.visitChildren(self)

    class TimeShiftExprContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def shift(self):
            return self.getTypedRuleContext(ExprParser.ShiftContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTimeShiftExpr"):
                return visitor.visitTimeShiftExpr(self)
            else:
                return visitor.visitChildren(self)

    class PortFieldContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def portFieldExpr(self):
            return self.getTypedRuleContext(ExprParser.PortFieldExprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPortField"):
                return visitor.visitPortField(self)
            else:
                return visitor.visitChildren(self)

    class MuldivContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.op = None  # Token
            self.copyFrom(ctx)

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ExprParser.ExprContext)
            else:
                return self.getTypedRuleContext(ExprParser.ExprContext, i)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitMuldiv"):
                return visitor.visitMuldiv(self)
            else:
                return visitor.visitChildren(self)

    class TimeSumContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.from_ = None  # ShiftContext
            self.to = None  # ShiftContext
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def shift(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ExprParser.ShiftContext)
            else:
                return self.getTypedRuleContext(ExprParser.ShiftContext, i)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTimeSum"):
                return visitor.visitTimeSum(self)
            else:
                return visitor.visitChildren(self)

    class TimeIndexContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def IDENTIFIER(self):
            return self.getToken(ExprParser.IDENTIFIER, 0)

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTimeIndex"):
                return visitor.visitTimeIndex(self)
            else:
                return visitor.visitChildren(self)

    class TimeShiftContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def IDENTIFIER(self):
            return self.getToken(ExprParser.IDENTIFIER, 0)

        def shift(self):
            return self.getTypedRuleContext(ExprParser.ShiftContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTimeShift"):
                return visitor.visitTimeShift(self)
            else:
                return visitor.visitChildren(self)

    class FunctionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def IDENTIFIER(self):
            return self.getToken(ExprParser.IDENTIFIER, 0)

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunction"):
                return visitor.visitFunction(self)
            else:
                return visitor.visitChildren(self)

    def expr(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = ExprParser.ExprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 4
        self.enterRecursionRule(localctx, 4, self.RULE_expr, _p)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 78
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 0, self._ctx)
            if la_ == 1:
                localctx = ExprParser.UnsignedAtomContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 22
                self.atom()
                pass

            elif la_ == 2:
                localctx = ExprParser.PortFieldContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 23
                self.portFieldExpr()
                pass

            elif la_ == 3:
                localctx = ExprParser.NegationContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 24
                self.match(ExprParser.T__1)
                self.state = 25
                self.expr(13)
                pass

            elif la_ == 4:
                localctx = ExprParser.ExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 26
                self.match(ExprParser.T__2)
                self.state = 27
                self.expr(0)
                self.state = 28
                self.match(ExprParser.T__3)
                pass

            elif la_ == 5:
                localctx = ExprParser.AllTimeSumContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 30
                self.match(ExprParser.T__7)
                self.state = 31
                self.match(ExprParser.T__2)
                self.state = 32
                self.expr(0)
                self.state = 33
                self.match(ExprParser.T__3)
                pass

            elif la_ == 6:
                localctx = ExprParser.PortFieldSumContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 35
                self.match(ExprParser.T__8)
                self.state = 36
                self.match(ExprParser.T__2)
                self.state = 37
                self.portFieldExpr()
                self.state = 38
                self.match(ExprParser.T__3)
                pass

            elif la_ == 7:
                localctx = ExprParser.TimeSumContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 40
                self.match(ExprParser.T__7)
                self.state = 41
                self.match(ExprParser.T__2)
                self.state = 42
                localctx.from_ = self.shift()
                self.state = 43
                self.match(ExprParser.T__9)
                self.state = 44
                localctx.to = self.shift()
                self.state = 45
                self.match(ExprParser.T__10)
                self.state = 46
                self.expr(0)
                self.state = 47
                self.match(ExprParser.T__3)
                pass

            elif la_ == 8:
                localctx = ExprParser.FunctionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 49
                self.match(ExprParser.IDENTIFIER)
                self.state = 50
                self.match(ExprParser.T__2)
                self.state = 51
                self.expr(0)
                self.state = 52
                self.match(ExprParser.T__3)
                pass

            elif la_ == 9:
                localctx = ExprParser.TimeShiftContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 54
                self.match(ExprParser.IDENTIFIER)
                self.state = 55
                self.match(ExprParser.T__11)
                self.state = 56
                self.shift()
                self.state = 57
                self.match(ExprParser.T__12)
                pass

            elif la_ == 10:
                localctx = ExprParser.TimeIndexContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 59
                self.match(ExprParser.IDENTIFIER)
                self.state = 60
                self.match(ExprParser.T__11)
                self.state = 61
                self.expr(0)
                self.state = 62
                self.match(ExprParser.T__12)
                pass

            elif la_ == 11:
                localctx = ExprParser.TimeShiftExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 64
                self.match(ExprParser.T__2)
                self.state = 65
                self.expr(0)
                self.state = 66
                self.match(ExprParser.T__3)
                self.state = 67
                self.match(ExprParser.T__11)
                self.state = 68
                self.shift()
                self.state = 69
                self.match(ExprParser.T__12)
                pass

            elif la_ == 12:
                localctx = ExprParser.TimeIndexExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 71
                self.match(ExprParser.T__2)
                self.state = 72
                self.expr(0)
                self.state = 73
                self.match(ExprParser.T__3)
                self.state = 74
                self.match(ExprParser.T__11)
                self.state = 75
                self.expr(0)
                self.state = 76
                self.match(ExprParser.T__12)
                pass

            self._ctx.stop = self._input.LT(-1)
            self.state = 91
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 2, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 89
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 1, self._ctx)
                    if la_ == 1:
                        localctx = ExprParser.MuldivContext(
                            self, ExprParser.ExprContext(self, _parentctx, _parentState)
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_expr
                        )
                        self.state = 80
                        if not self.precpred(self._ctx, 11):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 11)"
                            )
                        self.state = 81
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not (_la == 5 or _la == 6):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 82
                        self.expr(12)
                        pass

                    elif la_ == 2:
                        localctx = ExprParser.AddsubContext(
                            self, ExprParser.ExprContext(self, _parentctx, _parentState)
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_expr
                        )
                        self.state = 83
                        if not self.precpred(self._ctx, 10):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 10)"
                            )
                        self.state = 84
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not (_la == 2 or _la == 7):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 85
                        self.expr(11)
                        pass

                    elif la_ == 3:
                        localctx = ExprParser.ComparisonContext(
                            self, ExprParser.ExprContext(self, _parentctx, _parentState)
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_expr
                        )
                        self.state = 86
                        if not self.precpred(self._ctx, 9):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 9)"
                            )
                        self.state = 87
                        self.match(ExprParser.COMPARISON)
                        self.state = 88
                        self.expr(10)
                        pass

                self.state = 93
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 2, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class AtomContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ExprParser.RULE_atom

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class NumberContext(AtomContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.AtomContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NUMBER(self):
            return self.getToken(ExprParser.NUMBER, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitNumber"):
                return visitor.visitNumber(self)
            else:
                return visitor.visitChildren(self)

    class IdentifierContext(AtomContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.AtomContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def IDENTIFIER(self):
            return self.getToken(ExprParser.IDENTIFIER, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIdentifier"):
                return visitor.visitIdentifier(self)
            else:
                return visitor.visitChildren(self)

    def atom(self):
        localctx = ExprParser.AtomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_atom)
        try:
            self.state = 96
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [14]:
                localctx = ExprParser.NumberContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 94
                self.match(ExprParser.NUMBER)
                pass
            elif token in [16]:
                localctx = ExprParser.IdentifierContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 95
                self.match(ExprParser.IDENTIFIER)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ShiftContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TIME(self):
            return self.getToken(ExprParser.TIME, 0)

        def shift_expr(self):
            return self.getTypedRuleContext(ExprParser.Shift_exprContext, 0)

        def getRuleIndex(self):
            return ExprParser.RULE_shift

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitShift"):
                return visitor.visitShift(self)
            else:
                return visitor.visitChildren(self)

    def shift(self):
        localctx = ExprParser.ShiftContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_shift)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 98
            self.match(ExprParser.TIME)
            self.state = 100
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 2 or _la == 7:
                self.state = 99
                self.shift_expr(0)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Shift_exprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ExprParser.RULE_shift_expr

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class SignedAtomContext(Shift_exprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.Shift_exprContext
            super().__init__(parser)
            self.op = None  # Token
            self.copyFrom(ctx)

        def atom(self):
            return self.getTypedRuleContext(ExprParser.AtomContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSignedAtom"):
                return visitor.visitSignedAtom(self)
            else:
                return visitor.visitChildren(self)

    class SignedExpressionContext(Shift_exprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.Shift_exprContext
            super().__init__(parser)
            self.op = None  # Token
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSignedExpression"):
                return visitor.visitSignedExpression(self)
            else:
                return visitor.visitChildren(self)

    class ShiftMuldivContext(Shift_exprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.Shift_exprContext
            super().__init__(parser)
            self.op = None  # Token
            self.copyFrom(ctx)

        def shift_expr(self):
            return self.getTypedRuleContext(ExprParser.Shift_exprContext, 0)

        def right_expr(self):
            return self.getTypedRuleContext(ExprParser.Right_exprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitShiftMuldiv"):
                return visitor.visitShiftMuldiv(self)
            else:
                return visitor.visitChildren(self)

    class ShiftAddsubContext(Shift_exprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.Shift_exprContext
            super().__init__(parser)
            self.op = None  # Token
            self.copyFrom(ctx)

        def shift_expr(self):
            return self.getTypedRuleContext(ExprParser.Shift_exprContext, 0)

        def right_expr(self):
            return self.getTypedRuleContext(ExprParser.Right_exprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitShiftAddsub"):
                return visitor.visitShiftAddsub(self)
            else:
                return visitor.visitChildren(self)

    def shift_expr(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = ExprParser.Shift_exprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 10
        self.enterRecursionRule(localctx, 10, self.RULE_shift_expr, _p)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 110
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 5, self._ctx)
            if la_ == 1:
                localctx = ExprParser.SignedAtomContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 103
                localctx.op = self._input.LT(1)
                _la = self._input.LA(1)
                if not (_la == 2 or _la == 7):
                    localctx.op = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 104
                self.atom()
                pass

            elif la_ == 2:
                localctx = ExprParser.SignedExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 105
                localctx.op = self._input.LT(1)
                _la = self._input.LA(1)
                if not (_la == 2 or _la == 7):
                    localctx.op = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 106
                self.match(ExprParser.T__2)
                self.state = 107
                self.expr(0)
                self.state = 108
                self.match(ExprParser.T__3)
                pass

            self._ctx.stop = self._input.LT(-1)
            self.state = 120
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 118
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 6, self._ctx)
                    if la_ == 1:
                        localctx = ExprParser.ShiftMuldivContext(
                            self,
                            ExprParser.Shift_exprContext(
                                self, _parentctx, _parentState
                            ),
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_shift_expr
                        )
                        self.state = 112
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 4)"
                            )
                        self.state = 113
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not (_la == 5 or _la == 6):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 114
                        self.right_expr(0)
                        pass

                    elif la_ == 2:
                        localctx = ExprParser.ShiftAddsubContext(
                            self,
                            ExprParser.Shift_exprContext(
                                self, _parentctx, _parentState
                            ),
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_shift_expr
                        )
                        self.state = 115
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 3)"
                            )
                        self.state = 116
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not (_la == 2 or _la == 7):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 117
                        self.right_expr(0)
                        pass

                self.state = 122
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class Right_exprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ExprParser.RULE_right_expr

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class RightExpressionContext(Right_exprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.Right_exprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(ExprParser.ExprContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRightExpression"):
                return visitor.visitRightExpression(self)
            else:
                return visitor.visitChildren(self)

    class RightMuldivContext(Right_exprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.Right_exprContext
            super().__init__(parser)
            self.op = None  # Token
            self.copyFrom(ctx)

        def right_expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ExprParser.Right_exprContext)
            else:
                return self.getTypedRuleContext(ExprParser.Right_exprContext, i)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRightMuldiv"):
                return visitor.visitRightMuldiv(self)
            else:
                return visitor.visitChildren(self)

    class RightAtomContext(Right_exprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ExprParser.Right_exprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def atom(self):
            return self.getTypedRuleContext(ExprParser.AtomContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRightAtom"):
                return visitor.visitRightAtom(self)
            else:
                return visitor.visitChildren(self)

    def right_expr(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = ExprParser.Right_exprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 12
        self.enterRecursionRule(localctx, 12, self.RULE_right_expr, _p)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 129
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [3]:
                localctx = ExprParser.RightExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 124
                self.match(ExprParser.T__2)
                self.state = 125
                self.expr(0)
                self.state = 126
                self.match(ExprParser.T__3)
                pass
            elif token in [14, 16]:
                localctx = ExprParser.RightAtomContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 128
                self.atom()
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 136
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 9, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = ExprParser.RightMuldivContext(
                        self,
                        ExprParser.Right_exprContext(self, _parentctx, _parentState),
                    )
                    self.pushNewRecursionContext(
                        localctx, _startState, self.RULE_right_expr
                    )
                    self.state = 131
                    if not self.precpred(self._ctx, 3):
                        from antlr4.error.Errors import FailedPredicateException

                        raise FailedPredicateException(
                            self, "self.precpred(self._ctx, 3)"
                        )
                    self.state = 132
                    localctx.op = self._input.LT(1)
                    _la = self._input.LA(1)
                    if not (_la == 5 or _la == 6):
                        localctx.op = self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()
                    self.state = 133
                    self.right_expr(4)
                self.state = 138
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 9, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    def sempred(self, localctx: RuleContext, ruleIndex: int, predIndex: int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[2] = self.expr_sempred
        self._predicates[5] = self.shift_expr_sempred
        self._predicates[6] = self.right_expr_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def expr_sempred(self, localctx: ExprContext, predIndex: int):
        if predIndex == 0:
            return self.precpred(self._ctx, 11)

        if predIndex == 1:
            return self.precpred(self._ctx, 10)

        if predIndex == 2:
            return self.precpred(self._ctx, 9)

    def shift_expr_sempred(self, localctx: Shift_exprContext, predIndex: int):
        if predIndex == 3:
            return self.precpred(self._ctx, 4)

        if predIndex == 4:
            return self.precpred(self._ctx, 3)

    def right_expr_sempred(self, localctx: Right_exprContext, predIndex: int):
        if predIndex == 5:
            return self.precpred(self._ctx, 3)



from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Optional, Union

import pandas as pd
from attr import dataclass

# from gems.simulation.extra_output import ExtraOutput
# from optimization import OptimizationProblem  # Adjust import as needed
from gems.simulation.output_values import OutputValues


class SimulationColumns(str, Enum):
    BLOCK = "block"
    COMPONENT = "component"
    OUTPUT = "output"
    ABSOLUTE_TIME_INDEX = "absolute-time-index"
    BLOCK_TIME_INDEX = "block-time-index"
    SCENARIO_INDEX = "scenario-index"
    VALUE = "value"
    BASIS_STATUS = "basis-status"


class SimulationTableBuilder:
    """Builds simulation tables from solver output values."""

    def __init__(self, simulation_id: Optional[str] = None) -> None:
        self.simulation_id: str = simulation_id or datetime.now().strftime(
            "%Y%m%d-%H%M"
        )

    def build(
        self, output_values: OutputValues, absolute_time_offset: Optional[int] = None
    ) -> pd.DataFrame:
        if output_values.problem is None:
            raise ValueError("OutputValues problem is not set.")

        context = output_values.problem.context
        block = context._block.id
        block_size = context.block_length()

        absolute_time_offset = absolute_time_offset or (block - 1) * block_size
        assert absolute_time_offset is not None

        rows: list[dict[str, Any]] = []
        rows += self._collect_solver_outputs(output_values, block, absolute_time_offset)
        rows += self._collect_extra_outputs(output_values, block, absolute_time_offset)
        rows.append(self._collect_objective_value(output_values, block))

        return pd.DataFrame(rows, columns=[c.value for c in SimulationColumns])

    # -------------------------------------------------------------------------
    # Solver outputs
    # -------------------------------------------------------------------------
    def _collect_solver_outputs(
        self, output_values: OutputValues, block: int, abs_offset: int
    ) -> list[dict[str, Any]]:
        rows: list[dict[str, Any]] = []

        for comp_id, comp in output_values._components.items():
            if hasattr(comp, "_variables"):
                for var in comp._variables.values():
                    if hasattr(var, "_value"):
                        for ts_index, val in var._value.items():
                            basis_status = None
                            if hasattr(var, "_basis_status"):
                                if isinstance(var._basis_status, str):
                                    basis_status = var._basis_status
                                elif isinstance(var._basis_status, dict):
                                    basis_status = var._basis_status.get(ts_index)

                            var_name = var._name if hasattr(var, "_name") else ""
                            rows.append(
                                {
                                    SimulationColumns.BLOCK.value: block,
                                    SimulationColumns.COMPONENT.value: comp_id,
                                    SimulationColumns.OUTPUT.value: var_name,
                                    SimulationColumns.ABSOLUTE_TIME_INDEX.value: abs_offset
                                    + ts_index.time,
                                    SimulationColumns.BLOCK_TIME_INDEX.value: ts_index.time,
                                    SimulationColumns.SCENARIO_INDEX.value: ts_index.scenario,
                                    SimulationColumns.VALUE.value: val,
                                    SimulationColumns.BASIS_STATUS.value: basis_status,
                                }
                            )
        return rows

    # -------------------------------------------------------------------------
    # Extra outputs
    # -------------------------------------------------------------------------
    def _collect_extra_outputs(
        self, output_values: OutputValues, block: int, abs_offset: int
    ) -> list[dict[str, Any]]:
        rows: list[dict[str, Any]] = []

        for comp_id, comp in output_values._components.items():
            if hasattr(comp, "_extra_outputs"):
                for name, extra_output in comp._extra_outputs.items():
                    for ts_index, val in extra_output._value.items():
                        # if hasattr(extra_output, "values"):
                        #     for ts_index, val in extra_output.values.items():
                        t: int = ts_index.time
                        s: int = ts_index.scenario
                        rows.append(
                            {
                                SimulationColumns.BLOCK.value: block,
                                SimulationColumns.COMPONENT.value: comp_id,
                                SimulationColumns.OUTPUT.value: name,
                                SimulationColumns.ABSOLUTE_TIME_INDEX.value: abs_offset
                                + t,
                                SimulationColumns.BLOCK_TIME_INDEX.value: t,
                                SimulationColumns.SCENARIO_INDEX.value: s,
                                SimulationColumns.VALUE.value: val,
                                SimulationColumns.BASIS_STATUS.value: None,
                            }
                        )
        return rows

    # -------------------------------------------------------------------------
    # Objective value
    # -------------------------------------------------------------------------
    def _collect_objective_value(
        self, output_values: OutputValues, block: int
    ) -> dict[str, Any]:
        assert output_values.problem is not None, "OutputValues problem is not set"
        assert (
            output_values.problem.solver is not None
        ), "OutputValues problem.solver is not set"
        return {
            SimulationColumns.BLOCK.value: block,
            SimulationColumns.COMPONENT.value: None,
            SimulationColumns.OUTPUT.value: "objective-value",
            SimulationColumns.ABSOLUTE_TIME_INDEX.value: None,
            SimulationColumns.BLOCK_TIME_INDEX.value: None,
            SimulationColumns.SCENARIO_INDEX.value: None,
            SimulationColumns.VALUE.value: output_values.problem.solver.Objective().Value(),
            SimulationColumns.BASIS_STATUS.value: None,
        }


@dataclass
class SimulationTableWriter:
    """Handles writing simulation tables to CSV."""

    simulation_table: pd.DataFrame

    def write_csv(
        self,
        output_dir: Union[str, Path],
        simulation_id: str,
        optim_nb: int,
    ) -> Path:
        """Write the simulation table to CSV."""
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        filename = f"simulation_table_{simulation_id}_{optim_nb}.csv"
        filepath = output_dir / filename
        self.simulation_table.to_csv(filepath, index=False)
        return filepath



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
Module for common classes used in models.
"""
from enum import Enum


class ValueType(Enum):
    CONTINUOUS = "CONTINUOUS"
    INTEGER = "INTEGER"
    BOOLEAN = "BOOLEAN"


class ProblemContext(Enum):
    OPERATIONAL = 0
    INVESTMENT = 1
    COUPLING = 2



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass, replace
from typing import Any, Optional

from gems.expression import ExpressionNode, literal
from gems.expression.degree import is_constant
from gems.expression.equality import expressions_equal_if_present
from gems.expression.indexing_structure import IndexingStructure
from gems.model.common import ProblemContext, ValueType


@dataclass
class Variable:
    """
    A decision variable of the model.
    """

    name: str
    data_type: ValueType
    lower_bound: Optional[ExpressionNode]
    upper_bound: Optional[ExpressionNode]
    structure: IndexingStructure
    context: ProblemContext

    def __post_init__(self) -> None:
        if self.lower_bound and not is_constant(self.lower_bound):
            raise ValueError("Lower bounds of variables must be constant")
        if self.upper_bound and not is_constant(self.upper_bound):
            raise ValueError("Upper bounds of variables must be constant")

    def replicate(self, /, **changes: Any) -> "Variable":
        return replace(self, **changes)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Variable):
            return False
        return (
            self.name == other.name
            and self.data_type == other.data_type
            and expressions_equal_if_present(self.lower_bound, other.lower_bound)
            and expressions_equal_if_present(self.upper_bound, other.upper_bound)
            and self.structure == other.structure
        )


def int_variable(
    name: str,
    lower_bound: Optional[ExpressionNode] = None,
    upper_bound: Optional[ExpressionNode] = None,
    structure: IndexingStructure = IndexingStructure(True, True),
    context: ProblemContext = ProblemContext.OPERATIONAL,
) -> Variable:
    return Variable(
        name, ValueType.INTEGER, lower_bound, upper_bound, structure, context
    )


def bool_var(
    name: str,
    structure: IndexingStructure = IndexingStructure(True, True),
    context: ProblemContext = ProblemContext.OPERATIONAL,
) -> Variable:
    return Variable(name, ValueType.BOOLEAN, literal(0), literal(1), structure, context)


def float_variable(
    name: str,
    lower_bound: Optional[ExpressionNode] = None,
    upper_bound: Optional[ExpressionNode] = None,
    structure: IndexingStructure = IndexingStructure(True, True),
    context: ProblemContext = ProblemContext.OPERATIONAL,
) -> Variable:
    return Variable(
        name, ValueType.CONTINUOUS, lower_bound, upper_bound, structure, context
    )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
The model module defines the data model for user-defined models.
A model allows to define the behaviour for components, by
defining parameters, variables, and equations.
"""
import itertools
from dataclasses import dataclass, field, replace
from typing import Any, Dict, Iterable, Optional

from gems.expression import ExpressionNode
from gems.expression.degree import is_linear
from gems.expression.indexing import IndexingStructureProvider, compute_indexation
from gems.expression.indexing_structure import IndexingStructure
from gems.model.constraint import Constraint
from gems.model.parameter import Parameter
from gems.model.port import PortFieldDefinition, PortFieldId, PortType
from gems.model.variable import Variable


# TODO: Introduce bool_variable ?
def _make_structure_provider(model: "Model") -> IndexingStructureProvider:
    class Provider(IndexingStructureProvider):
        def get_parameter_structure(self, name: str) -> IndexingStructure:
            return model.parameters[name].structure

        def get_variable_structure(self, name: str) -> IndexingStructure:
            return model.variables[name].structure

        def get_component_parameter_structure(
            self, component_id: str, name: str
        ) -> IndexingStructure:
            raise NotImplementedError(
                "Cannot have parameters associated to components in models."
            )

        def get_component_variable_structure(
            self, component_id: str, name: str
        ) -> IndexingStructure:
            raise NotImplementedError(
                "Cannot have variables associated to components in models."
            )

    return Provider()


def _is_objective_contribution_valid(
    model: "Model", objective_contribution: ExpressionNode
) -> bool:
    if not is_linear(objective_contribution):
        raise ValueError("Objective contribution must be a linear expression.")

    data_structure_provider = _make_structure_provider(model)
    objective_structure = compute_indexation(
        objective_contribution, data_structure_provider
    )

    if objective_structure != IndexingStructure(time=False, scenario=False):
        raise ValueError("Objective contribution should be a real-valued expression.")
    # TODO: We should also check that the number of instances is equal to 1, but this would require a linearization here, do not want to do that for now...
    return True


@dataclass(frozen=True)
class ModelPort:
    """
    Instance of a port as a model member.

    A model may carry multiple ports of the same type.
    For example, the 2 ports at line extremities.
    """

    port_type: PortType
    port_name: str

    def replicate(self, /, **changes: Any) -> "ModelPort":
        return replace(self, **changes)


@dataclass(frozen=True)
class Model:
    """
    Defines a model that can be referenced by actual components.
    A model defines the behaviour of those components.
    """

    id: str
    constraints: Dict[str, Constraint] = field(default_factory=dict)
    binding_constraints: Dict[str, Constraint] = field(default_factory=dict)
    inter_block_dyn: bool = False
    parameters: Dict[str, Parameter] = field(default_factory=dict)
    variables: Dict[str, Variable] = field(default_factory=dict)
    objective_contributions: Optional[Dict[str, ExpressionNode]] = None
    objective_operational_contribution: Optional[ExpressionNode] = None
    objective_investment_contribution: Optional[ExpressionNode] = None
    ports: Dict[str, ModelPort] = field(default_factory=dict)
    port_fields_definitions: Dict[PortFieldId, PortFieldDefinition] = field(
        default_factory=dict
    )
    extra_outputs: Optional[Dict[str, ExpressionNode]] = None

    def __post_init__(self) -> None:
        if self.objective_operational_contribution:
            _is_objective_contribution_valid(
                self, self.objective_operational_contribution
            )

        if self.objective_investment_contribution:
            _is_objective_contribution_valid(
                self, self.objective_investment_contribution
            )
        # Validate each contribution if present
        if self.objective_contributions:
            for expr in self.objective_contributions.values():
                _is_objective_contribution_valid(self, expr)

        for definition in self.port_fields_definitions.values():
            port_name = definition.port_field.port_name
            port_field = definition.port_field.field_name
            port = self.ports.get(port_name, None)
            if port is None:
                raise ValueError(f"Invalid port in port field definition: {port_name}")
            if port_field not in [f.name for f in port.port_type.fields]:
                raise ValueError(
                    f"Invalid port field in port field definition: {port_field}"
                )

    def get_all_constraints(self) -> Iterable[Constraint]:
        """
        Get binding constraints and inner constraints altogether.
        """
        return itertools.chain(
            self.binding_constraints.values(), self.constraints.values()
        )

    def replicate(self, /, **changes: Any) -> "Model":
        # Shallow copy
        return replace(self, **changes)


def model(
    id: str,
    constraints: Optional[Iterable[Constraint]] = None,
    binding_constraints: Optional[Iterable[Constraint]] = None,
    parameters: Optional[Iterable[Parameter]] = None,
    variables: Optional[Iterable[Variable]] = None,
    objective_contributions: Optional[Dict[str, ExpressionNode]] = None,
    objective_operational_contribution: Optional[ExpressionNode] = None,
    objective_investment_contribution: Optional[ExpressionNode] = None,
    inter_block_dyn: bool = False,
    ports: Optional[Iterable[ModelPort]] = None,
    port_fields_definitions: Optional[Iterable[PortFieldDefinition]] = None,
    extra_outputs: Optional[Dict[str, ExpressionNode]] = None,
) -> Model:
    """
    Utility method to create Models from relaxed arguments
    """
    existing_port_names = {}
    if ports:
        for port in ports:
            port_name = port.port_name
            if port_name not in existing_port_names:
                existing_port_names[port_name] = port
            else:
                raise ValueError(
                    f"2 ports have the same name inside the model, it's not authorized : {port_name}"
                )
    return Model(
        id=id,
        constraints={c.name: c for c in constraints} if constraints else {},
        binding_constraints={c.name: c for c in binding_constraints}
        if binding_constraints
        else {},
        parameters={p.name: p for p in parameters} if parameters else {},
        variables={v.name: v for v in variables} if variables else {},
        objective_contributions=objective_contributions,
        objective_investment_contribution=objective_investment_contribution,
        objective_operational_contribution=objective_operational_contribution,
        inter_block_dyn=inter_block_dyn,
        ports=existing_port_names,
        port_fields_definitions={d.port_field: d for d in port_fields_definitions}
        if port_fields_definitions
        else {},
        extra_outputs=extra_outputs,
    )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

import pandas as pd

from gems.model import Model
from gems.model.library import Library
from gems.study import (
    Component,
    ConstantData,
    DataBase,
    Network,
    Node,
    PortRef,
    PortsConnection,
    Scenarization,
)
from gems.study.data import (
    AbstractDataStructure,
    ScenarioSeriesData,
    TimeScenarioSeriesData,
    TimeSeriesData,
    dataframe_to_scenario_series,
    dataframe_to_time_series,
    load_ts_from_file,
)
from gems.study.parsing import InputComponent, InputPortConnections, InputSystem


@dataclass(frozen=True)
class System:
    components: Dict[str, Component]
    nodes: Dict[str, Component]
    connections: List[PortsConnection]


def system(
    components_list: Iterable[Component],
    nodes: Iterable[Component],
    connections: Iterable[PortsConnection],
) -> System:
    return System(
        components=dict((m.id, m) for m in components_list),
        nodes=dict((n.id, n) for n in nodes),
        connections=list(connections),
    )


def resolve_system(input_system: InputSystem, libraries: dict[str, Library]) -> System:
    """
    Resolves:
    - components to be used for study
    - connections between components"""
    components_list = [
        _resolve_component(libraries, m) for m in input_system.components
    ]
    nodes_input = getattr(input_system, "nodes", []) or []
    nodes = [_resolve_component(libraries, n) for n in nodes_input]
    all_components: List[Component] = components_list + nodes
    connections_input = getattr(input_system, "connections", []) or []
    connections = []
    for cnx in connections_input:
        resolved_cnx = _resolve_connections(cnx, all_components)
        connections.append(resolved_cnx)

    return system(components_list, nodes, connections)


def _resolve_component(
    libraries: dict[str, Library], component: InputComponent
) -> Component:
    lib_id, model_id = component.model.split(".")
    model = libraries[lib_id].models[model_id]

    return Component(
        model=model,
        id=component.id,
    )


def _resolve_connections(
    connection: InputPortConnections,
    all_components: List[Component],
) -> PortsConnection:
    cnx_component1 = connection.component1
    cnx_component2 = connection.component2
    port1 = connection.port1
    port2 = connection.port2

    component_1 = _get_component_by_id(all_components, cnx_component1)
    component_2 = _get_component_by_id(all_components, cnx_component2)
    assert component_1 is not None and component_2 is not None
    port_ref_1 = PortRef(component_1, port1)
    port_ref_2 = PortRef(component_2, port2)

    return PortsConnection(port_ref_1, port_ref_2)


def _get_component_by_id(
    all_components: List[Component], component_id: str
) -> Optional[Component]:
    components_dict = {component.id: component for component in all_components}
    return components_dict.get(component_id)


def consistency_check(
    input_study: Dict[str, Component], input_models: Dict[str, Model]
) -> bool:
    """
    Checks if all components in the Components instances have a valid model from the library.
    Returns True if all components are consistent, raises ValueError otherwise.
    """
    # TODO: Update this consistency check to check if each component have a valid model from the lib it refers to (and not all libs)
    model_ids_set = input_models.keys()
    for component_id, component in input_study.items():
        if component.model.id not in model_ids_set:
            raise ValueError(
                f"Error: Component {component_id} has invalid model ID: {component.model.id}"
            )
    return True


def build_network(system: System) -> Network:
    # It seems that System and Network are almost the same thing -> could be simplified ?
    network = Network("study")

    for node_id, node in system.nodes.items():
        node = Node(model=node.model, id=node_id)
        network.add_node(node)

    for component in system.components.values():
        network.add_component(component)

    for connection in system.connections:
        network.connect(connection.port1, connection.port2)
    return network


def build_data_base(
    input_system: InputSystem, timeseries_dir: Optional[Path]
) -> DataBase:
    database = DataBase()
    nodes_input = getattr(input_system, "nodes", []) or []
    input_system_objects = input_system.components + nodes_input
    for comp in input_system_objects:
        # This idiom allows mypy to 'ignore' the fact that comp.parameter can be None
        for param in comp.parameters or []:
            param_value = _build_data(
                param.time_dependent,
                param.scenario_dependent,
                param.value,
                timeseries_dir,
            )
            database.add_data(comp.id, param.id, param_value)

    return database


def _build_data(
    time_dependent: bool,
    scenario_dependent: bool,
    param_value: Union[float, str],
    timeseries_dir: Optional[Path],
    scenarization: Optional[Scenarization] = None,
) -> AbstractDataStructure:
    if isinstance(param_value, str):
        # Should happen only if time-dependent or scenario-dependent
        ts_data = load_ts_from_file(param_value, timeseries_dir)
        if time_dependent and scenario_dependent:
            return TimeScenarioSeriesData(ts_data, scenarization)
        elif time_dependent:
            return TimeSeriesData(dataframe_to_time_series(ts_data))
        elif scenario_dependent:
            return ScenarioSeriesData(
                dataframe_to_scenario_series(ts_data), scenarization
            )
        else:
            raise ValueError(
                f"A float value is expected for constant data, got {param_value}"
            )
    else:  # param_value is a float, we should be in the constant data case
        if time_dependent or scenario_dependent:
            raise ValueError(
                f"A timeseries name is expected for time or scenario dependent data, got {param_value}"
            )
        return ConstantData(float(param_value))


def _resolve_scenarization(
    scenario_builder_data: pd.DataFrame,
) -> Dict[str, Scenarization]:
    output: Dict[str, Scenarization] = {}
    for i, row in scenario_builder_data.iterrows():
        if row["name"] in output:
            output[row["name"]].add_year(row["year"], row["scenario"])
        else:
            output[row["name"]] = Scenarization({row["year"]: row["scenario"]})
    return output


def build_scenarized_data_base(
    input_comp: InputSystem,
    scenario_builder_data: pd.DataFrame,
    timeseries_dir: Optional[Path],
) -> DataBase:
    database = DataBase()
    scenarizations = _resolve_scenarization(scenario_builder_data)

    for comp in input_comp.components:
        scenarization = None
        if comp.scenario_group:
            scenarization = scenarizations[comp.scenario_group]

        # This idiom allows mypy to 'ignore' the fact that comp.parameter can be None
        for param in comp.parameters or []:
            if param.scenario_group:
                scenarization = scenarizations[param.scenario_group]
            param_value = _build_data(
                param.time_dependent,
                param.scenario_dependent,
                param.value,
                timeseries_dir,
                scenarization,
            )
            database.add_data(comp.id, param.id, param_value)

    return database



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
The optimization module contains the logic to translate the input model
into a mathematical optimization problem.
"""

import itertools
import math
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Iterable, List, Optional

import ortools.linear_solver.pywraplp as lp

from gems.expression import EvaluationVisitor, ExpressionNode, ValueProvider, visit
from gems.expression.context_adder import add_component_context
from gems.expression.indexing import IndexingStructureProvider, compute_indexation
from gems.expression.indexing_structure import IndexingStructure
from gems.expression.operators_expansion import ProblemDimensions, expand_operators
from gems.expression.port_resolver import PortFieldKey, resolve_port
from gems.model.common import ValueType
from gems.model.constraint import Constraint
from gems.model.port import PortFieldId
from gems.simulation.linear_expression import LinearExpression, Term
from gems.simulation.linearize import ParameterGetter, linearize_expression
from gems.simulation.strategy import (
    MergedProblemStrategy,
    ModelSelectionStrategy,
    RiskManagementStrategy,
    UniformRisk,
)
from gems.simulation.time_block import TimeBlock
from gems.study.data import DataBase
from gems.study.network import Component, Network
from gems.utils import get_or_add


@dataclass(eq=True, frozen=True)
class TimestepComponentVariableKey:
    """
    Identifies the solver variable for one timestep and one component variable.
    """

    component_id: str
    variable_name: str
    block_timestep: Optional[int] = None
    scenario: Optional[int] = None


def _get_parameter_value(
    context: "OptimizationContext",
    block_timestep: Optional[int],
    scenario: Optional[int],
    component_id: str,
    name: str,
) -> float:
    data = context.database.get_data(component_id, name)
    absolute_timestep = context.block_timestep_to_absolute_timestep(block_timestep)
    return data.get_value(absolute_timestep, scenario, context.tree_node)


def _make_value_provider(
    context: "OptimizationContext",
    block_timestep: Optional[int],
    scenario: Optional[int],
) -> ValueProvider:
    """
    Create a value provider which takes its values from
    the parameter values as defined in the network data.

    Cannot evaluate expressions which contain variables.
    """

    class Impl(ValueProvider):
        def get_component_variable_value(self, component_id: str, name: str) -> float:
            raise NotImplementedError(
                "Cannot provide variable value at problem build time."
            )

        def get_component_parameter_value(self, component_id: str, name: str) -> float:
            return _get_parameter_value(
                context, block_timestep, scenario, component_id, name
            )

        def get_variable_value(self, name: str) -> float:
            raise NotImplementedError(
                "Cannot provide variable value at problem build time."
            )

        def get_parameter_value(self, name: str) -> float:
            raise ValueError(
                "Parameter must be associated to its component before resolution."
            )

    return Impl()


def _compute_expression_value(
    expression: ExpressionNode,
    context: "OptimizationContext",
    block_timestep: Optional[int],
    scenario: Optional[int],
) -> float:
    value_provider = _make_value_provider(context, block_timestep, scenario)
    visitor = EvaluationVisitor(value_provider)
    return visit(expression, visitor)


class BlockBorderManagement(Enum):
    """
    Class to specify the way of handling the time horizon (or time block) border.
        - IGNORE_OUT_OF_FRAME: Ignore terms in constraints that lead to out of horizon data
        - CYCLE: Consider all timesteps to be specified modulo the horizon length, this is the actual functioning of Antares
    """

    IGNORE_OUT_OF_FRAME = "IGNORE"
    CYCLE = "CYCLE"


@dataclass
class SolverVariableInfo:
    """
    Helper class for constructing the structure file
    for Benders solver. It keeps track of the corresponding
    column of the variable in the MPS format as well as if it is
    present in the objective function or not
    """

    name: str
    column_id: int
    is_in_objective: bool


def float_to_int(value: float) -> int:
    if isinstance(value, int) or value.is_integer():
        return int(value)
    else:
        raise ValueError(f"{value} is not an integer.")


class OptimizationContext:
    """
    Helper class to build the optimization problem.

     - Maintains mappings between model and solver objects,
     - Maintains mappings between port fields and expressions,
     - Provides implementations of interfaces required by various visitors
       used to transform expressions (values providers ...).
    """

    def __init__(
        self,
        network: Network,
        database: DataBase,
        block: TimeBlock,
        scenarios: int,
        border_management: BlockBorderManagement,
        build_strategy: ModelSelectionStrategy = MergedProblemStrategy(),
        risk_strategy: RiskManagementStrategy = UniformRisk(),
        decision_tree_node: str = "",
        use_full_var_name: bool = True,
    ):
        self._network = network
        self._database = database
        self._block = block
        self._scenarios = scenarios
        self._border_management = border_management
        self._build_strategy = build_strategy
        self._risk_strategy = risk_strategy
        self._tree_node = decision_tree_node
        self._full_var_name = use_full_var_name

        self._component_variables: Dict[TimestepComponentVariableKey, lp.Variable] = {}
        self._solver_variables: Dict[str, SolverVariableInfo] = {}
        self._connection_fields_expressions: Dict[
            PortFieldKey, List[ExpressionNode]
        ] = {}

        self._constant_value_provider = self._make_constant_value_provider()
        self._indexing_structure_provider = self._make_data_structure_provider()
        self._parameter_getter = self._make_parameter_getter()

    @property
    def network(self) -> Network:
        return self._network

    @property
    def scenarios(self) -> int:
        return self._scenarios

    @property
    def tree_node(self) -> str:
        return self._tree_node

    @property
    def build_strategy(self) -> ModelSelectionStrategy:
        return self._build_strategy

    @property
    def risk_strategy(self) -> RiskManagementStrategy:
        return self._risk_strategy

    @property
    def full_var_name(self) -> bool:
        return self._full_var_name

    def block_length(self) -> int:
        return len(self._block.timesteps)

    @property
    def connection_fields_expressions(self) -> Dict[PortFieldKey, List[ExpressionNode]]:
        return self._connection_fields_expressions

    # TODO: Need to think about data processing when creating blocks with varying or inequal time steps length (aggregation, sum ?, mean of data ?)
    def block_timestep_to_absolute_timestep(
        self, block_timestep: Optional[int]
    ) -> Optional[int]:
        """
        Timestep may be None for parameters or variables that don't depend on time.
        """
        if block_timestep is None:
            return None
        return self._block.timesteps[self.get_actual_block_timestep(block_timestep)]

    def get_actual_block_timestep(self, block_timestep: int) -> int:
        if self._border_management == BlockBorderManagement.CYCLE:
            return block_timestep % self.block_length()
        else:
            raise NotImplementedError()

    @property
    def database(self) -> DataBase:
        return self._database

    def _manage_border_timesteps(self, timestep: int) -> int:
        if self._border_management == BlockBorderManagement.CYCLE:
            return timestep % self.block_length()
        else:
            raise NotImplementedError

    def get_time_indices(self, index_structure: IndexingStructure) -> Iterable[int]:
        return range(self.block_length()) if index_structure.time else range(1)

    def get_scenario_indices(self, index_structure: IndexingStructure) -> Iterable[int]:
        return range(self.scenarios) if index_structure.scenario else range(1)

    def get_component_variable(
        self,
        block_timestep: Optional[int],
        scenario: Optional[int],
        component_id: str,
        variable_name: str,
    ) -> lp.Variable:
        if block_timestep is not None:
            block_timestep = self._manage_border_timesteps(block_timestep)
        return self._component_variables[
            TimestepComponentVariableKey(
                component_id, variable_name, block_timestep, scenario
            )
        ]

    def get_all_component_variables(
        self,
    ) -> Dict[TimestepComponentVariableKey, lp.Variable]:
        return self._component_variables

    def register_component_variable(
        self,
        block_timestep: Optional[int],
        scenario: Optional[int],
        component_id: str,
        model_var_name: str,
        variable: lp.Variable,
    ) -> None:
        key = TimestepComponentVariableKey(
            component_id, model_var_name, block_timestep, scenario
        )
        if key not in self._component_variables:
            self._solver_variables[variable.name()] = SolverVariableInfo(
                variable.name(), len(self._solver_variables), False
            )
        self._component_variables[key] = variable

    def register_connection_fields_expressions(
        self,
        component_id: str,
        port_name: str,
        field_name: str,
        expression: ExpressionNode,
    ) -> None:
        key = PortFieldKey(component_id, PortFieldId(port_name, field_name))
        get_or_add(self._connection_fields_expressions, key, lambda: []).append(
            expression
        )

    def evaluate_time_bound(self, expression: ExpressionNode) -> int:
        res = visit(expression, EvaluationVisitor(self._constant_value_provider))
        return float_to_int(res)

    def _make_data_structure_provider(self) -> IndexingStructureProvider:
        """
        Retrieve information in data structure (parameter and variable) from the model
        """
        network = self.network

        class Impl(IndexingStructureProvider):
            def get_component_variable_structure(
                self, component_id: str, name: str
            ) -> IndexingStructure:
                return (
                    network.get_component(component_id).model.variables[name].structure
                )

            def get_component_parameter_structure(
                self, component_id: str, name: str
            ) -> IndexingStructure:
                return (
                    network.get_component(component_id).model.parameters[name].structure
                )

            def get_parameter_structure(self, name: str) -> IndexingStructure:
                raise RuntimeError("Component context should have been initialized.")

            def get_variable_structure(self, name: str) -> IndexingStructure:
                raise RuntimeError("Component context should have been initialized.")

        return Impl()

    def expand_operators(self, expression: ExpressionNode) -> ExpressionNode:
        dimensions = ProblemDimensions(self.block_length(), self.scenarios)
        time_bound_evaluator = self.evaluate_time_bound
        return expand_operators(
            expression,
            dimensions,
            time_bound_evaluator,
            self._indexing_structure_provider,
        )

    def _make_parameter_getter(self) -> ParameterGetter:
        ctxt = self

        class Impl(ParameterGetter):
            def get_parameter_value(
                self,
                component_id: str,
                parameter_name: str,
                timestep: Optional[int],
                scenario: Optional[int],
            ) -> float:
                return _get_parameter_value(
                    ctxt,
                    timestep,
                    scenario,
                    component_id,
                    parameter_name,
                )

        return Impl()

    def linearize_expression(
        self,
        expanded: ExpressionNode,
        timestep: Optional[int] = None,
        scenario: Optional[int] = None,
    ) -> LinearExpression:
        return linearize_expression(
            expanded, timestep, scenario, self._parameter_getter
        )

    def compute_indexing(self, expression: ExpressionNode) -> IndexingStructure:
        return compute_indexation(expression, self._indexing_structure_provider)

    def _make_constant_value_provider(self) -> ValueProvider:
        """
        Value provider which only provides values for constant parameters
        """
        context = self
        network = self.network

        class Impl(ValueProvider):
            def get_component_variable_value(
                self, component_id: str, name: str
            ) -> float:
                raise NotImplementedError(
                    "Cannot provide variable value at problem build time."
                )

            def get_component_parameter_value(
                self, component_id: str, name: str
            ) -> float:
                model = network.get_component(component_id).model
                structure = model.parameters[name].structure
                if structure.time or structure.scenario:
                    raise ValueError(f"Parameter {name} is not constant.")
                return _get_parameter_value(context, None, None, component_id, name)

            def get_variable_value(self, name: str) -> float:
                raise NotImplementedError(
                    "Cannot provide variable value at problem build time."
                )

            def get_parameter_value(self, name: str) -> float:
                raise ValueError(
                    "Parameter must be associated to its component before resolution."
                )

        return Impl()


def _compute_indexing(
    context: OptimizationContext, constraint: Constraint
) -> IndexingStructure:
    return (
        context.compute_indexing(constraint.expression)
        or context.compute_indexing(constraint.lower_bound)
        or context.compute_indexing(constraint.upper_bound)
    )


def _instantiate_model_expression(
    model_expression: ExpressionNode,
    component_id: str,
    optimization_context: OptimizationContext,
) -> ExpressionNode:
    """
    Performs common operations that are necessary on model expressions before their actual use:
     1. add component ID for variables and parameters of THIS component
     2. replace port fields by their definition
    """
    with_component = add_component_context(component_id, model_expression)
    with_component_and_ports = resolve_port(
        with_component, component_id, optimization_context.connection_fields_expressions
    )
    return with_component_and_ports


def _create_constraint(
    solver: lp.Solver,
    context: OptimizationContext,
    constraint: Constraint,
) -> None:
    """
    Adds a component-related constraint to the solver.
    """
    expanded = context.expand_operators(constraint.expression)
    constraint_indexing = _compute_indexing(context, constraint)

    for block_timestep in context.get_time_indices(constraint_indexing):
        for scenario in context.get_scenario_indices(constraint_indexing):
            linear_expr_at_t = context.linearize_expression(
                expanded, block_timestep, scenario
            )

            # What happens if there is some time_operator in the bounds ?
            constraint_data = ConstraintData(
                name=constraint.name,
                lower_bound=_compute_expression_value(
                    constraint.lower_bound,
                    context,
                    block_timestep,
                    scenario,
                ),
                upper_bound=_compute_expression_value(
                    constraint.upper_bound,
                    context,
                    block_timestep,
                    scenario,
                ),
                expression=linear_expr_at_t,
            )
            make_constraint(
                solver,
                context,
                block_timestep,
                scenario,
                constraint_data,
            )


def _create_objective(
    solver: lp.Solver,
    opt_context: OptimizationContext,
    component: Component,
    objective_contribution: ExpressionNode,
) -> None:
    instantiated_expr = _instantiate_model_expression(
        objective_contribution, component.id, opt_context
    )
    expanded = opt_context.expand_operators(instantiated_expr)
    linear_expr = opt_context.linearize_expression(expanded)

    _add_linear_expression_to_objective(solver, opt_context, linear_expr)


def _add_linear_expression_to_objective(
    solver: lp.Solver,
    context: OptimizationContext,
    linear_expr: LinearExpression,
    existing_obj: Optional[lp.Objective] = None,
) -> lp.Objective:
    """
    Adds terms from a LinearExpression to an existing lp.Objective, or creates a new one.
    This function handles the accumulation logic required for objective contributions.
    """
    obj: lp.Objective = existing_obj if existing_obj is not None else solver.Objective()

    for term in linear_expr.terms.values():
        solver_var = _get_solver_var(term, context)
        context._solver_variables[solver_var.name()].is_in_objective = True
        obj.SetCoefficient(
            solver_var,
            obj.GetCoefficient(solver_var) + term.coefficient,
        )

    obj.SetOffset(linear_expr.constant + obj.offset())
    return obj


@dataclass
class ConstraintData:
    name: str
    lower_bound: float
    upper_bound: float
    expression: LinearExpression


def _get_solver_var(
    term: Term,
    context: OptimizationContext,
) -> lp.Variable:
    return context.get_component_variable(
        term.time_index,
        term.scenario_index,
        term.component_id,
        term.variable_name,
    )


def make_constraint(
    solver: lp.Solver,
    context: OptimizationContext,
    block_timestep: int,
    scenario: int,
    data: ConstraintData,
) -> None:
    """
    Adds constraint to the solver.
    """
    constraint_name = _solver_constraint_name(context, block_timestep, scenario, data)

    solver_constraint: lp.Constraint = solver.Constraint(constraint_name)
    constant: float = 0
    for term in data.expression.terms.values():
        solver_var = _get_solver_var(
            term,
            context,
        )
        coefficient = term.coefficient + solver_constraint.GetCoefficient(solver_var)
        solver_constraint.SetCoefficient(solver_var, coefficient)

    constant += data.expression.constant
    solver_constraint.SetBounds(
        data.lower_bound - constant, data.upper_bound - constant
    )


def _solver_constraint_name(
    context: OptimizationContext,
    block_timestep: int,
    scenario: int,
    data: ConstraintData,
) -> str:
    scenario_suffix = (
        f"_s{scenario}" if (scenario is not None and context.scenarios > 1) else ""
    )
    block_suffix = (
        f"_t{block_timestep}"
        if (block_timestep is not None and context.block_length() > 1)
        else ""
    )
    constraint_name = f"{data.name}{scenario_suffix}{block_suffix}"
    return constraint_name


class OptimizationProblem:
    name: str
    solver: lp.Solver
    context: OptimizationContext

    def __init__(
        self,
        name: str,
        solver: lp.Solver,
        opt_context: OptimizationContext,
    ) -> None:
        self.name = name
        self.solver = solver
        self.context = opt_context

        self._register_connection_fields_definitions()
        self._create_variables()
        self._create_constraints()
        self._create_objectives()

    def _register_connection_fields_definitions(self) -> None:
        for cnx in self.context.network.connections:
            for field_name in list(cnx.master_port.keys()):
                master_port = cnx.master_port[field_name]
                port_definition = (
                    master_port.component.model.port_fields_definitions.get(
                        PortFieldId(
                            port_name=master_port.port_id, field_name=field_name.name
                        )
                    )
                )
                expression_node = port_definition.definition  # type: ignore
                instantiated_expression = add_component_context(
                    master_port.component.id, expression_node
                )
                self.context.register_connection_fields_expressions(
                    component_id=cnx.port1.component.id,
                    port_name=cnx.port1.port_id,
                    field_name=field_name.name,
                    expression=instantiated_expression,
                )
                self.context.register_connection_fields_expressions(
                    component_id=cnx.port2.component.id,
                    port_name=cnx.port2.port_id,
                    field_name=field_name.name,
                    expression=instantiated_expression,
                )

    def _solver_variable_name(
        self, component_id: str, var_name: str, t: Optional[int], s: Optional[int]
    ) -> str:
        component_prefix = (
            f"{component_id}." if (self.context.full_var_name and component_id) else ""
        )
        tree_prefix = (
            f"{self.context.tree_node}."
            if (self.context.full_var_name and self.context.tree_node)
            else ""
        )
        scenario_suffix = (
            f"_s{s}" if (s is not None and self.context.scenarios > 1) else ""
        )
        block_suffix = (
            f"_t{t}" if (t is not None and self.context.block_length() > 1) else ""
        )

        # Set solver var name
        # Externally, for the Solver, this variable will have a full name
        # Internally, it will be indexed by a structure that into account
        # the component id, variable name, timestep and scenario separately
        return (
            f"{tree_prefix}{component_prefix}{var_name}{scenario_suffix}{block_suffix}"
        )

    def _create_variables(self) -> None:
        for component in self.context.network.all_components:
            model = component.model

            for model_var in self.context.build_strategy.get_variables(model):
                var_indexing = model_var.structure
                instantiated_lb_expr = None
                instantiated_ub_expr = None

                if model_var.lower_bound:
                    instantiated_lb_expr = _instantiate_model_expression(
                        model_var.lower_bound, component.id, self.context
                    )
                if model_var.upper_bound:
                    instantiated_ub_expr = _instantiate_model_expression(
                        model_var.upper_bound, component.id, self.context
                    )

                time_indices: Iterable[Optional[int]] = [None]
                if var_indexing.is_time_varying():
                    time_indices = self.context.get_time_indices(var_indexing)

                scenario_indices: Iterable[Optional[int]] = [None]
                if var_indexing.is_scenario_varying():
                    scenario_indices = self.context.get_scenario_indices(var_indexing)

                for t, s in itertools.product(time_indices, scenario_indices):
                    lower_bound = -self.solver.infinity()
                    upper_bound = self.solver.infinity()
                    if instantiated_lb_expr:
                        lower_bound = _compute_expression_value(
                            instantiated_lb_expr, self.context, t, s
                        )
                    if instantiated_ub_expr:
                        upper_bound = _compute_expression_value(
                            instantiated_ub_expr, self.context, t, s
                        )

                    solver_var_name = self._solver_variable_name(
                        component.id, model_var.name, t, s
                    )

                    if lower_bound > upper_bound:
                        raise ValueError(
                            f"Upper bound ({upper_bound}) must be strictly greater than lower bound ({lower_bound}) for variable {solver_var_name}"
                        )

                    if model_var.data_type == ValueType.BOOLEAN:
                        solver_var = self.solver.BoolVar(
                            solver_var_name,
                        )
                    elif model_var.data_type == ValueType.INTEGER:
                        solver_var = self.solver.IntVar(
                            lower_bound,
                            upper_bound,
                            solver_var_name,
                        )
                    else:
                        solver_var = self.solver.NumVar(
                            lower_bound,
                            upper_bound,
                            solver_var_name,
                        )
                    self.context.register_component_variable(
                        t, s, component.id, model_var.name, solver_var
                    )

    def _create_constraints(self) -> None:
        for component in self.context.network.all_components:
            for constraint in self.context.build_strategy.get_constraints(
                component.model
            ):
                instantiated_expr = _instantiate_model_expression(
                    constraint.expression, component.id, self.context
                )
                instantiated_lb = _instantiate_model_expression(
                    constraint.lower_bound, component.id, self.context
                )
                instantiated_ub = _instantiate_model_expression(
                    constraint.upper_bound, component.id, self.context
                )

                instantiated_constraint = Constraint(
                    name=f"{component.id}.{constraint.name}",
                    expression=instantiated_expr,
                    lower_bound=instantiated_lb,
                    upper_bound=instantiated_ub,
                )
                _create_constraint(
                    self.solver,
                    self.context,
                    instantiated_constraint,
                )

    def _create_objectives(self) -> None:
        """
        Iterates over all network components and creates their corresponding
        optimization objectives in the solver.
        """
        for component in self.context.network.all_components:
            model = component.model
            # We distinguish between objective_contributions and classical objective retreived from context build_strategy
            if getattr(model, "objective_contributions", None):
                self._create_objectives_from_contributions(component)
            else:
                self._create_objectives_from_build_strategy(component)

    def _create_objectives_from_build_strategy(self, component: "Component") -> None:
        model = component.model
        for objective in self.context.build_strategy.get_objectives(model):
            if objective is not None:
                _create_objective(
                    self.solver,
                    self.context,
                    component,
                    self.context.risk_strategy(objective),
                )

    def _create_objectives_from_contributions(self, component: "Component") -> None:
        """
        Creates objectives from the 'objective_contributions' dictionary defined
        directly on the component's model (new format).
        """
        model = component.model
        if model.objective_contributions is None:
            return

        main_objective: Optional[lp.Objective] = None
        for contrib_id, expr in model.objective_contributions.items():
            if expr is None:
                continue

            instantiated = _instantiate_model_expression(
                expr, component.id, self.context
            )
            expanded = self.context.expand_operators(instantiated)
            linear_expr = self.context.linearize_expression(expanded)

            main_objective = _add_linear_expression_to_objective(
                self.solver, self.context, linear_expr, existing_obj=main_objective
            )

    def export_as_mps(self) -> str:
        return self.solver.ExportModelAsMpsFormat(fixed_format=True, obfuscated=False)

    def export_as_lp(self) -> str:
        return self.solver.ExportModelAsLpFormat(obfuscated=False)


def build_problem(
    network: Network,
    database: DataBase,
    block: TimeBlock,
    scenarios: int,
    *,
    problem_name: str = "optimization_problem",
    border_management: BlockBorderManagement = BlockBorderManagement.CYCLE,
    solver_id: str = "SCIP",
    build_strategy: ModelSelectionStrategy = MergedProblemStrategy(),
    risk_strategy: RiskManagementStrategy = UniformRisk(),
    decision_tree_node: str = "",
    use_full_var_name: bool = True,
) -> OptimizationProblem:
    """
    Entry point to build the optimization problem for a time period.
    """
    solver: lp.Solver = lp.Solver.CreateSolver(solver_id)

    database.requirements_consistency(network)

    opt_context = OptimizationContext(
        network,
        database,
        block,
        scenarios,
        border_management,
        build_strategy,
        risk_strategy,
        decision_tree_node,
        use_full_var_name,
    )

    return OptimizationProblem(problem_name, solver, opt_context)


def fusion_problems(
    masters: List[OptimizationProblem], coupler: OptimizationProblem
) -> OptimizationProblem:
    if len(masters) == 1:
        # Nothing to fusion. Just past down the master
        return masters[0]

    root_master = coupler
    root_master.name = "master"

    root_vars: Dict[str, lp.Variable] = dict()
    root_constraints: Dict[str, lp.Constraint] = dict()
    root_objective = root_master.solver.Objective()

    # We stock the coupler's variables to check for
    # same name variables in the masters
    for var in root_master.solver.variables():
        root_vars[var.name()] = var

    for master in masters:
        context = master.context
        objective = master.solver.Objective()

        for var in master.solver.variables():
            # If variable not already in coupler, we add it
            # Otherwise we update its upper and lower bounds
            if var.name() not in root_vars:
                root_var = root_master.solver.NumVar(var.lb(), var.ub(), var.name())
                root_master.context._solver_variables[var.name()] = SolverVariableInfo(
                    var.name(),
                    len(root_master.context._solver_variables),
                    context._solver_variables[var.name()].is_in_objective,
                )
            else:
                root_var = root_vars[var.name()]
                root_var.SetLb(var.lb())
                root_var.SetUb(var.ub())
                root_master.context._solver_variables[
                    var.name()
                ].is_in_objective = context._solver_variables[
                    var.name()
                ].is_in_objective

            for cstr in master.solver.constraints():
                coeff = cstr.GetCoefficient(var)
                # If variable present in constraint, we add the constraint to root
                if coeff != 0:
                    key = f"{master.name}.{cstr.name()}"
                    if key not in root_constraints:
                        root_constraints[key] = root_master.solver.Constraint(
                            cstr.Lb(), cstr.Ub(), key
                        )
                    root_cstr = root_constraints[key]
                    root_cstr.SetCoefficient(root_var, coeff)

            obj_coeff = objective.GetCoefficient(var)
            if obj_coeff != 0:
                root_objective.SetCoefficient(root_var, obj_coeff)

    return root_master



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
from typing import Dict, List, Optional, Set

from gems.expression import ExpressionNode, literal
from gems.expression.indexing_structure import IndexingStructure
from gems.expression.parsing.parse_expression import ModelIdentifiers, parse_expression
from gems.model import (
    Constraint,
    Model,
    ModelPort,
    Parameter,
    PortField,
    PortType,
    ProblemContext,
    ValueType,
    Variable,
    model,
)
from gems.model.library import Library
from gems.model.model import model
from gems.model.parsing import (
    InputConstraint,
    InputField,
    InputLibrary,
    InputModel,
    InputModelPort,
    InputParameter,
    InputPortFieldDefinition,
    InputPortType,
    InputVariable,
)
from gems.model.port import PortFieldDefinition, port_field_def


def resolve_library(
    input_libs: List[InputLibrary], preloaded_libs: Optional[List[Library]] = None
) -> Dict[str, Library]:
    """
    Converts parsed data into an actually usable library of models.

     - resolves references between models and ports
     - parses expressions and resolves references to variables/params
    """
    yaml_lib_dict = dict((l.id, l) for l in input_libs)

    preloaded_port_types = {}
    if preloaded_libs:
        for preloaded_lib in preloaded_libs:
            preloaded_port_types.update(preloaded_lib.port_types)

    output_lib_dict: Dict[str, Library] = (
        dict((l.id, l) for l in preloaded_libs) if preloaded_libs else {}
    )

    remaining_lib_ids: List[str] = list(yaml_lib_dict)
    treated_lib_ids: Set[str] = set()
    import_stack: List[str] = []

    while remaining_lib_ids:
        next_lib_id = remaining_lib_ids.pop()

        if next_lib_id in treated_lib_ids:
            continue
        else:
            import_stack.append(next_lib_id)

        while import_stack:
            cur_yaml_lib = yaml_lib_dict[import_stack[-1]]
            current_lib = Library(id=cur_yaml_lib.id, port_types={}, models={})

            # Add already parsed port types from dependencies in current lib
            _add_preloaded_port_types_to_current_lib(preloaded_port_types, current_lib)
            _add_resolved_dependent_port_types_to_current_lib(
                output_lib_dict, treated_lib_ids, cur_yaml_lib, current_lib
            )

            remaining_dependencies = set(cur_yaml_lib.dependencies) - treated_lib_ids

            if remaining_dependencies:
                _add_dependencies_to_stack(import_stack, remaining_dependencies)

            else:
                _resolve_lib(current_lib, cur_yaml_lib, output_lib_dict)
                _update_treated_libs_and_import_stack(treated_lib_ids, import_stack)

    return output_lib_dict


def _add_preloaded_port_types_to_current_lib(
    preloaded_port_types: dict[str, PortType], current_lib: Library
) -> None:
    current_lib.port_types.update(preloaded_port_types)


def _add_resolved_dependent_port_types_to_current_lib(
    output_lib_dict: Dict[str, Library],
    treated_lib_ids: Set[str],
    cur_yaml_lib: InputLibrary,
    current_lib: Library,
) -> None:
    done_dependencies = set(cur_yaml_lib.dependencies) & treated_lib_ids
    for done_lib in done_dependencies:
        current_lib.port_types.update(output_lib_dict[done_lib].port_types)


def _update_treated_libs_and_import_stack(
    treated_lib_ids: Set[str], import_stack: List[str]
) -> None:
    treated_lib_ids.add(import_stack.pop())


def _resolve_lib(
    current_lib: Library, cur_yaml_lib: InputLibrary, output_lib: Dict[str, Library]
) -> None:
    port_types = [_convert_port_type(p) for p in cur_yaml_lib.port_types]
    port_types_dict = dict((p.id, p) for p in port_types)

    if current_lib.port_types.keys() & port_types_dict.keys():
        raise Exception(
            f"Port(s) : {str(current_lib.port_types.keys() & port_types_dict.keys())} is(are) defined twice."
        )
    current_lib.port_types.update(port_types_dict)

    cur_yaml_lib_model_ids = [model.id for model in cur_yaml_lib.models]
    for id in cur_yaml_lib_model_ids:
        if cur_yaml_lib_model_ids.count(id) > 1:
            raise Exception(f"Model {id} is defined twice")

    models = [_resolve_model(m, current_lib.port_types) for m in cur_yaml_lib.models]

    models_dict = dict((m.id, m) for m in models)

    current_lib.models.update(models_dict)
    output_lib[current_lib.id] = current_lib


def _add_dependencies_to_stack(
    import_stack: List[str], remaining_dependencies: Set[str]
) -> None:
    first_dependency = remaining_dependencies.pop()

    if first_dependency in import_stack:
        raise Exception("Circular import in yaml libraries")
    import_stack.append(first_dependency)


def _convert_field(field: InputField) -> PortField:
    return PortField(name=field.id)


def _convert_port_type(port_type: InputPortType) -> PortType:
    return PortType(
        id=port_type.id, fields=[_convert_field(f) for f in port_type.fields]
    )


def _resolve_model(input_model: InputModel, port_types: Dict[str, PortType]) -> Model:
    identifiers = ModelIdentifiers(
        variables={v.id for v in input_model.variables},
        parameters={p.id for p in input_model.parameters},
    )

    objective_contributions = None
    if input_model.objective_contributions:
        objective_contributions = {
            contrib.id: parse_expression(contrib.expression, identifiers)
            for contrib in input_model.objective_contributions
        }

    extra_outputs = (
        {
            eo.id: parse_expression(eo.expression, identifiers)
            for eo in input_model.extra_outputs
        }
        if input_model.extra_outputs
        else None
    )
    return model(
        id=input_model.id,
        parameters=[_to_parameter(p) for p in input_model.parameters],
        variables=[_to_variable(v, identifiers) for v in input_model.variables],
        ports=[_resolve_model_port(p, port_types) for p in input_model.ports],
        port_fields_definitions=[
            _resolve_field_definition(d, identifiers)
            for d in input_model.port_field_definitions
        ],
        binding_constraints=[
            _to_constraint(c, identifiers) for c in input_model.binding_constraints
        ],
        constraints=[_to_constraint(c, identifiers) for c in input_model.constraints],
        objective_contributions=objective_contributions,
        extra_outputs=extra_outputs,
    )


def _resolve_model_port(
    port: InputModelPort, port_types: Dict[str, PortType]
) -> ModelPort:
    return ModelPort(port_name=port.id, port_type=port_types[port.type])


def _resolve_field_definition(
    definition: InputPortFieldDefinition, ids: ModelIdentifiers
) -> PortFieldDefinition:
    return port_field_def(
        port_name=definition.port,
        field_name=definition.field,
        definition=parse_expression(definition.definition, ids),
    )


def _to_parameter(param: InputParameter) -> Parameter:
    return Parameter(
        name=param.id,
        type=ValueType.CONTINUOUS,
        structure=IndexingStructure(param.time_dependent, param.scenario_dependent),
    )


def _to_expression_if_present(
    expr: Optional[str], identifiers: ModelIdentifiers
) -> Optional[ExpressionNode]:
    if not expr:
        return None
    return parse_expression(expr, identifiers)


def _to_variable(var: InputVariable, identifiers: ModelIdentifiers) -> Variable:
    return Variable(
        name=var.id,
        data_type={"continuous": ValueType.CONTINUOUS, "integer": ValueType.INTEGER}[
            var.variable_type
        ],
        structure=IndexingStructure(var.time_dependent, var.scenario_dependent),
        lower_bound=_to_expression_if_present(var.lower_bound, identifiers),
        upper_bound=_to_expression_if_present(var.upper_bound, identifiers),
        context=ProblemContext.OPERATIONAL,
    )


def _to_constraint(
    constraint: InputConstraint, identifiers: ModelIdentifiers
) -> Constraint:
    lb = _to_expression_if_present(constraint.lower_bound, identifiers)
    ub = _to_expression_if_present(constraint.upper_bound, identifiers)
    return Constraint(
        name=constraint.id,
        expression=parse_expression(constraint.expression, identifiers),
        lower_bound=(lb if lb is not None else literal(-float("inf"))),
        upper_bound=(ub if ub is not None else literal(float("inf"))),
    )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass

from . import CopyVisitor
from .expression import (
    ComponentParameterNode,
    ComponentVariableNode,
    ExpressionNode,
    ParameterNode,
    ProblemParameterNode,
    ProblemVariableNode,
    VariableNode,
)
from .visitor import visit


@dataclass(frozen=True)
class ContextAdder(CopyVisitor):
    """
    Simply copies the whole AST but associates all variables and parameters
    to the provided component ID.
    """

    component_id: str

    def variable(self, node: VariableNode) -> ExpressionNode:
        return ComponentVariableNode(self.component_id, node.name)

    def parameter(self, node: ParameterNode) -> ExpressionNode:
        return ComponentParameterNode(self.component_id, node.name)

    def comp_variable(self, node: ComponentVariableNode) -> ExpressionNode:
        raise ValueError(
            "This expression has already been associated to another component."
        )

    def comp_parameter(self, node: ComponentParameterNode) -> ExpressionNode:
        raise ValueError(
            "This expression has already been associated to another component."
        )

    def pb_variable(self, node: ProblemVariableNode) -> ExpressionNode:
        raise ValueError(
            "This expression has already been associated to another component."
        )

    def pb_parameter(self, node: ProblemParameterNode) -> ExpressionNode:
        raise ValueError(
            "This expression has already been associated to another component."
        )


def add_component_context(id: str, expression: ExpressionNode) -> ExpressionNode:
    return visit(expression, ContextAdder(id))
